# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-
"""Module containing classes for DID definitions.

This module is used to parse DID definition files and merge with the unit definitions to find DIDs in a project.
It then generates the DID definition c-files for the supplier DID API.
"""

import csv
import os
from pathlib import Path
from ruamel.yaml import YAML
from powertrain_build import build_defs
from powertrain_build.lib.helper_functions import deep_dict_update
from powertrain_build.problem_logger import ProblemLogger
from powertrain_build.types import byte_size, get_ec_type, get_float32_types
from powertrain_build.unit_configs import CodeGenerators


def get_dids_in_prj(unit_cfgs):
    """Return a dict with DIDs in the currently included SW-Units.

    Args:
        unit_cfgs (UnitConfigs): Unit definitions.
    Returns:
        error_message (str): Message in case something went wrong.
        dict: a dict with all dids in the project, in the below format:

    ::

        {'DID_VARIABLE_NAME': {
            'handle': 'VcRegCh/VcRegCh/Subsystem/VcRegCh/1000_VcRegCh/1600_DID/Gain14',
            'configs': ['all'],
            'type': 'UInt32',
            'unit': '-',
            'lsb': 1,
            'max': 20,
            'min': 0,
            'offset': 0,
            'description': 'Actual Regen State',
            'name': 'DID_VARIABLE_NAME',
            'class': 'CVC_DISP'
            }
        }
    """
    dids_prj = {}
    error_messages = []
    unit_cfg = unit_cfgs.get_per_unit_cfg()
    for unit, data in unit_cfg.items():
        dids = data.get('dids')
        if dids is None:
            error_messages.append(f'No "dids" key in unit config for {unit}.')
            continue
        for name, did in dids.items():
            dids_prj[name] = did
    return error_messages, dids_prj


class DIDs(ProblemLogger):
    """A class for handling of DID definitions."""

    def __init__(self, build_cfg, unit_cfgs):
        """Parse DID definition files referenced by project config.

        Args:
            build_cfg (BuildProjConfig): Project configuration
            unit_cfgs (UnitConfigs): Unit definitions
        """
        super().__init__()
        self._build_cfg = build_cfg
        self._unit_cfgs = unit_cfgs
        did_filename = self._build_cfg.get_did_cfg_file_name()
        cfg_dir = self._build_cfg.get_prj_cfg_dir()
        did_f32_cfg_file = os.path.join(cfg_dir, did_filename + '_Float32.csv')
        did_u32_cfg_file = os.path.join(cfg_dir, did_filename + '_UInt32.csv')
        self._dids_f32 = self._load_did_config_files(did_f32_cfg_file)
        self._dids_u32 = self._load_did_config_files(did_u32_cfg_file)
        self.fh_h = None
        self.fh_c = None
        get_did_error_messages, self._did_dict = get_dids_in_prj(unit_cfgs)
        self._did_defs = self.get_did_config()
        self._float32_types = get_float32_types()
        if get_did_error_messages:
            self.critical('\n'.join(get_did_error_messages))

    def _load_did_config_files(self, config_file):
        """Load the did config files."""
        dids = {}
        with open(config_file, mode='r', encoding='utf-8') as did_fh:
            csv_did = csv.reader(did_fh, delimiter=';')
            did = list(csv_did)
            dids['dids'] = {row[0]: int(row[1], 16) for row in did[3:]}
            dids['start_did'] = int(did[1][0], 16)
            dids['end_did'] = int(did[1][1], 16)
        self._check_dids(dids)
        return dids

    @staticmethod
    def _check_dids(dids):
        """Check that all dids are within the start and end values."""
        start_did = dids['start_did']
        end_did = dids['end_did']

        for var, did in dids['dids'].items():
            if did < start_did:
                raise ValueError(f'{var} has a too low did 0x{did:X} start did is 0x{start_did:X}')
            if did > end_did:
                raise ValueError(f'{var} has a too high did 0x{did:X} start did is 0x{start_did:X}')

    def gen_did_def_files(self, filename):
        """Generate the VcDidDefinitions.c & h files used by the Did-API."""
        with open(filename + '.h', 'w', encoding="utf-8") as self.fh_h:
            with open(filename + '.c', 'w', encoding="utf-8") as self.fh_c:
                dids_f32, dids_u32, errors = self._check_and_reformat_dids()
                self._gen_did_def_c_file(dids_f32, dids_u32, errors)
                self._gen_did_def_h_file(dids_f32, dids_u32)
        return errors

    def _check_and_reformat_dids(self):
        """Check that DIDs are defined and create two new dicts."""
        dids_f32 = {}
        dids_u32 = {}
        did_def_f32s = self._did_defs['Float32']['dids']
        did_def_u32s = self._did_defs['UInt32']['dids']
        errors = []
        for sig in sorted(self._did_dict.keys()):
            did = self._did_dict[sig]
            if did['type'] in self._float32_types:
                if sig in did_def_f32s:
                    dids_f32[did_def_f32s[sig]] = did
                else:
                    msg = f'Did for Float32 signal "{sig}" not defined'
                    self.critical(msg)
                    errors.append(msg)
            else:
                if sig in did_def_u32s:
                    dids_u32[did_def_u32s[sig]] = did
                else:
                    msg = f'Did for UInt32 signal "{sig}" not defined'
                    self.critical(msg)
                    errors.append(msg)
        return (dids_f32, dids_u32, errors)

    def _get_datatypes(self):
        tl_types = ['UInt8', 'Int8', 'UInt16', 'Int16', 'UInt32', 'Int32', 'Float32', 'Bool']
        data_types_tl = [f'{tl_type}_' for tl_type in tl_types]
        data_types_ec = [f'{get_ec_type(tl_type)}_' for tl_type in tl_types]
        if len(self._unit_cfgs.code_generators) > 1:
            self.warning('Cannot generate DIDs for more than one generator.'
                         'Defaulting to TargetLink')
            return ', '.join(data_types_tl)
        if CodeGenerators.target_link in self._unit_cfgs.code_generators:
            return ', '.join(data_types_tl)
        return ', '.join(data_types_ec)

    def _get_type(self, tl_type):
        if CodeGenerators.target_link in self._unit_cfgs.code_generators:
            return tl_type
        return get_ec_type(tl_type)

    def _gen_did_def_h_file(self, dids_f32, dids_u32):
        """Generate the VcDidDefinitions.h files used by the Did-API."""
        _, f_name = os.path.split(self.fh_h.name)
        header_def_name = f_name.upper().replace('.', '_')
        self.fh_h.write(f'#ifndef {header_def_name}\n')
        self.fh_h.write(f'#define {header_def_name}\n\n')
        self.fh_h.write(self._unit_cfgs.base_types_headers)
        self.fh_h.write(f'enum Datatypes {{{self._get_datatypes()}}};\n\n')
        self.fh_h.write(f'#define DID_DATASTRUCT_LEN_FLOAT32 {len(dids_f32)}\n')
        self.fh_h.write(f'#define DID_DATASTRUCT_LEN_UINT32 {len(dids_u32)}\n\n')
        uint16_type = self._get_type('UInt16')
        float32_type = self._get_type('Float32')
        self.fh_h.write('struct DID_Mapping_UInt32 {\n\t'
                        f'{uint16_type} DID;'
                        '\n\tvoid* data;\n\tenum Datatypes type;\n};\n\n')
        self.fh_h.write('struct DID_Mapping_Float32 {\n\t'
                        f'{uint16_type} DID;'
                        '\n\t'
                        f'{float32_type}* data;'
                        '\n};\n\n')

        self.fh_h.write(f'#include "{build_defs.PREDECL_START}"\n')

        self.fh_h.write('extern const struct DID_Mapping_UInt32 DID_data_struct_UInt32[];\n')
        self.fh_h.write('extern const struct DID_Mapping_Float32 DID_data_struct_Float32[];\n')

        self.fh_h.write('/* Floats */\n')

        for key in sorted(dids_f32.keys()):
            did = dids_f32[key]
            self.fh_h.write(f'extern {did["type"]} {did["name"]}; /* Did id: 0x{key:X} */\n')

        self.fh_h.write('/* Integers & Bools */\n')

        for key in sorted(dids_u32.keys()):
            did = dids_u32[key]
            self.fh_h.write(f'extern {did["type"]} {did["name"]}; /* Did id: 0x{key:X} */\n')

        self.fh_h.write(f'#include "{build_defs.PREDECL_END}"\n')
        self.fh_h.write(f'\n#endif /* {header_def_name} */\n')

    def _gen_did_def_c_file(self, dids_f32, dids_u32, errors):
        """Generate the VcDidDefinitions.c files used by the Did-API."""
        _, filename = os.path.split(self.fh_h.name)
        self.fh_c.write(f'#include "{filename}"\n\n')
        self.fh_c.write(f'#include "{build_defs.CVC_CODE_START}"\n\n')
        self.fh_c.write('/* The table shall be sorted in ascending Did is order!\n'
                        ' If not the search algorithm does not work */\n')
        self.fh_c.write('const struct DID_Mapping_Float32 DID_data_struct_Float32[] = {\n')

        keys = sorted(dids_f32.keys())
        for key in keys:
            did = dids_f32[key]
            if key == keys[-1]:
                delim = ' '
            else:
                delim = ','
            self.fh_c.write('\t{0x%X, &%s}%c /* %s */ \n' %
                            (key, did['name'], delim, did['handle']))
        if not keys:
            self.fh_c.write('\t{0x0000, 0L} /* Dummy entry */ \n')
        self.fh_c.write('};\n\n')

        self.fh_c.write('const struct DID_Mapping_UInt32 DID_data_struct_UInt32[] = {\n')
        keys = sorted(dids_u32.keys())
        for key in keys:
            did = dids_u32[key]
            if key == keys[-1]:
                delim = ' '
            else:
                delim = ','
            self.fh_c.write('\t{0x%X, &%s, %s_}%c /* %s */ \n' %
                            (key, did['name'], did['type'], delim, did['handle']))

        if not keys:
            self.fh_c.write(f'\t{{0x0000, 0L, {self._get_type("UInt32")}_}} /* Dummy entry */ \n')
        self.fh_c.write('};\n\n')

        if errors:
            self.fh_c.write('/* *** DIDs not in the definition file! ****\n')
            for error in errors:
                self.fh_c.write(f'{error}\n')
            self.fh_c.write('*/\n')

        self.fh_c.write(f'\n#include "{build_defs.CVC_CODE_END}"\n')
        self.fh_c.write('\n/*------------------------------------------------------'
                        '----------------------*\\\n  END OF FILE\n\\*-------------'
                        '---------------------------------------------------------------*/')

    def gen_did_carcom_extract(self, filename):
        """Generate the csv-file used for carcom database import."""
        with open(filename, 'w', encoding="utf-8") as carcom_file:
            for sig in sorted(self._did_dict.keys()):
                did = self._did_dict[sig]
                carcom_file.write(self._format_did_csv_line(did))

    @staticmethod
    def _convert_value(value, type, default_value=0):
        if value in ['', '-']:
            return type(default_value)
        return type(value)

    @staticmethod
    def _hex_location(value):
        return hex(value).upper().lstrip('0X')

    def _format_did_csv_line(self, did):
        """Format the line based on the did.

        Arguments:
        did (dict): DID data
        """
        did_line = '{' + '};{'.join(['location',
                                     'description',
                                     'name',
                                     'name',
                                     'bytes',
                                     'offset',
                                     'bits',
                                     'data_type',
                                     'nine',
                                     'ten',
                                     'low',
                                     'high',
                                     'scaling',
                                     'compare',
                                     'unit',
                                     'sixteen',
                                     'service',
                                     'eighteen',
                                     'sessions']) + '}\n'
        float_format = '06'
        compare = ''
        did_bytes = 4
        did_offset = 0  # Always use 0. Not sure why.
        did_bits = 8 * did_bytes
        service = 17
        sessions = '22: 01 02 03'
        unknown = ''  # Fields were empty in old system
        did_def_f32s = self._did_defs['Float32']['dids']
        did_def_u32s = self._did_defs['UInt32']['dids']
        if did['name'] in did_def_f32s:
            location = self._hex_location(did_def_f32s[did['name']])
        elif did['name'] in did_def_u32s:
            location = self._hex_location(did_def_u32s[did['name']])
        else:
            self.warning('Could not find location for %s', did['name'])
            location = unknown
        if did['type'] in self._float32_types:
            did_type = '4-byte float'
            scaling = 'x*1'
        else:
            did_type = 'Unsigned'
            u32_scaling_base = '(x-2147483647){{operator}}{{lsb:{float_format}}} {{sign}} {{offset:{float_format}}}'
            u32_scaling = u32_scaling_base.format(float_format=float_format)
            offset = self._convert_value(did['offset'], float, 0)
            if offset > 0:
                sign = '+'
            else:
                sign = '-'
            lsb = self._convert_value(did['lsb'], float, 1)
            if lsb > 0:
                operator = '*'
            else:
                operator = '/'
                lsb = 1.0/lsb  # Why we do this, I do not know.
            scaling = u32_scaling.format(operator=operator,
                                         lsb=lsb,
                                         sign=sign,
                                         offset=offset)

        return did_line.format(location=location,
                               name=did['name'],
                               description=did['description'],
                               bytes=did_bytes,
                               offset=did_offset,
                               bits=did_bits,
                               data_type=did_type,
                               nine=unknown,
                               ten=unknown,
                               low=did['min'],
                               high=did['max'],
                               scaling=scaling,
                               compare=compare,
                               unit=did['unit'],
                               sixteen=unknown,
                               service=service,
                               eighteen=unknown,
                               sessions=sessions)

    def get_did_config(self):
        """Return a dict with the defined DIDs for all configs.

        Returns:
            dict: a dict with the DIDs defined for all configs

        """
        # self._checkConfig()
        return {'Float32': self._dids_f32, 'UInt32': self._dids_u32}


class HIDIDs(ProblemLogger):
    """A class for handling of HI DID definitions."""

    def __init__(self, build_cfg, unit_cfgs):
        """Init.

        Args:
            build_cfg (BuildProjConfig): Project configuration
            unit_cfgs (UnitConfigs): Unit definitions
        """
        super().__init__()
        self._build_cfg = build_cfg
        self._unit_cfgs = unit_cfgs
        self.file_name = 'VcDIDAPI'
        self.did_dict = self._compose_did_data()

    def _load_did_config_files(self, config_file):
        """Load the did config files.

        Args:
            config_file (str): Path to DID configuration file.
        Returns:
            dids (dict): Parsed DIDs from the configuration file.
        """
        dids = {}
        config_file_path = Path(config_file)
        if config_file_path.exists():
            with config_file_path.open(mode='r', encoding='utf-8') as did_fh:
                yaml = YAML(typ='safe', pure=True)
                dids = self._verify_did_config_dict(yaml.load(did_fh))
        else:
            self.warning(f'Unable to parse DIDs. Cannot find file: {config_file_path.as_posix()}.')
        return dids

    def _verify_did_config_dict(self, dids):
        """Verify the structure of the dict from the DID configuration file.
        Missing keys will be added but also produce critical errors.

        Args:
            dids (dict): DIDs parsed from DID configuration file.
        Returns:
            (dict): Updated DID dict.
        """
        optional_keys = {
            'nr_of_bytes',
        }
        expected_keys = {
            'id',
            'data_type',
            'function_type',
        }
        expected_function_type_keys = {
            'read_data',
            'read_data_max',
            'read_data_min',
            'condition_check',
            'condition_check_max',
            'condition_check_min',
        }
        for did, did_data in dids.items():
            did_keys = set(did_data.keys())
            used_optional_keys = did_keys & optional_keys
            unknown_keys = did_keys - (expected_keys | optional_keys)
            missing_keys = expected_keys - did_keys
            for key in used_optional_keys:
                self.info(f'Using optional key {key} for DID {did}.')
            for key in unknown_keys:
                self.warning(f'Ignoring unknown element {key} for DID {did}.')
                del did_data[key]
            for key in missing_keys:
                self.critical(f'DID {did} is missing element {key}.')
                did_data[key] = '<missing>'
            if did_data['function_type'] not in expected_function_type_keys:
                self.critical(f"DID {did} lists unknown function type {did_data['function_type']}")
                did_data['function_type'] = '<missing>'
        return dids

    def _compose_did_data(self):
        """Gather and merge DID data from project simulink models and DID configuration file.

        Returns:
            did_dict (dict): Dict containing project DID data.
        """
        get_did_error_messages, project_dids = get_dids_in_prj(self._unit_cfgs)
        if get_did_error_messages:
            self.critical('\n'.join(get_did_error_messages))
            return {}

        did_filename = self._build_cfg.get_did_cfg_file_name()
        config_directory = self._build_cfg.get_prj_cfg_dir()
        did_config_file = os.path.join(config_directory, did_filename)
        dids = self._load_did_config_files(did_config_file)

        did_dict = self.verify_dids(project_dids, dids)
        for data in did_dict.values():
            data['function'] = self.compose_did_function(data)

        return did_dict

    @staticmethod
    def compose_did_function(did_data):
        """Compose DID function calls.
        Args:
            did_data (dict): Dict describing a DID in the project.
        Returns:
            function (str): Function to generate for given DID.
        """
        did_id = did_data["id"]
        data_type = did_data["data_type"]
        type_to_function_map = {
            '<missing>': f'DID_{did_id}_Missing({data_type} *Data)',
            'read_data': f'DID_{did_id}_Runnable_ReadData({data_type} *Data)',
            'read_data_max': f'DID_{did_id}_Runnable_MAX_ReadData({data_type} *Data)',
            'read_data_min': f'DID_{did_id}_Runnable_MIN_ReadData({data_type} *Data)',
            'condition_check': f'DID_{did_id}_Runnable_ConditionCheckRead({data_type} *ErrorCode)',
            'condition_check_max': f'DID_{did_id}_Runnable_MAX_ConditionCheckRead({data_type} *ErrorCode)',
            'condition_check_min': f'DID_{did_id}_Runnable_MIN_ConditionCheckRead({data_type} *ErrorCode)'
        }
        return type_to_function_map[did_data['function_type']]

    def verify_dids(self, project_dids, dids):
        """Verify the DIDs.

        * Model DIDs must be defined in DID configuration file.
        * ID numbers can only appear once per function type.

        Args:
            project_dids (dict): DIDs listed in project/simulink models.
            dids (dict): DIDs listed in the DID configuration file.
        Returns:
            valid_dids (dict): Validated DIDs listed in both DID configuration file as well as project.
        """
        valid_dids = {}
        did_id_usage = {}

        if not project_dids:
            for did in dids:
                self.warning(f'Ignoring DID {did}, not defined in any model.')
            return valid_dids

        for name in project_dids:
            if name not in dids:
                self.warning(f'DID {name} not defined in DID defintion file.')
                continue

            did_id = dids[name]['id']
            function_type = dids[name]['function_type']
            if did_id in did_id_usage:
                if function_type in did_id_usage[did_id]:
                    self.critical(
                        f'ID {did_id} is '
                        f'already used for DID {did_id_usage[did_id][function_type]} of '
                        f'function type {function_type}.'
                    )
                    continue
                did_id_usage[did_id][function_type] = name
            else:
                did_id_usage[did_id] = {function_type: name}

            valid_dids[name] = deep_dict_update(dids[name], project_dids[name])

        return valid_dids

    def get_header_file_content(self):
        """Get content for the DID API header file.

        Returns:
            (list(str)): List of lines to write to DID API header file.
        """
        name = self._build_cfg.get_a2l_cfg()['name']
        header_guard = f'{self.file_name.upper()}_H'
        header = [
            f'#ifndef {header_guard}\n',
            f'#define {header_guard}\n',
            '\n',
            '#include "tl_basetypes.h"\n',
            f'#include "Rte_{name}.h"\n',
            '\n'
        ]
        footer = [f'\n#endif /* {header_guard} */\n']

        if not self.did_dict:
            return header + footer

        body = [f'#include "{build_defs.PREDECL_DISP_ASIL_D_START}"\n']
        for did_data in self.did_dict.values():
            define = did_data["class"].split('/')[-1]  # E.q. for ASIL D it is ASIL_D/CVC_DISP_ASIL_D
            body.append(f'extern {define} {did_data["type"]} {did_data["name"]};\n')
        body.append(f'#include "{build_defs.PREDECL_DISP_ASIL_D_END}"\n')

        body.append(f'\n#include "{build_defs.PREDECL_CODE_ASIL_D_START}"\n')
        for did_data in self.did_dict.values():
            body.append(f'void {did_data["function"]};\n')
        body.append(f'#include "{build_defs.PREDECL_CODE_ASIL_D_END}"\n')

        return header + body + footer

    def get_source_file_content(self):
        """Get content for the DID API source file.

        Returns:
            (list(str)): List of lines to write to DID API source file.
        """
        header = [
            f'#include "{self.file_name}.h"\n',
            '\n'
        ]

        if not self.did_dict:
            return header

        body = [f'#include "{build_defs.CVC_CODE_ASIL_D_START}"\n']
        for did, did_data in self.did_dict.items():
            size = f'{did_data["nr_of_bytes"]}' if 'nr_of_bytes' in did_data else f'sizeof({did_data["data_type"]})'
            if 'ConditionCheckRead' in did_data["function"]:
                argument = 'ErrorCode'
            else:
                argument = 'Data'
            body.extend([
                f'void {did_data["function"]}\n',
                '{\n',
                f'    memcpy({argument}, &{did}, {size});\n',
                '}\n'
            ])
        body.append(f'#include "{build_defs.CVC_CODE_ASIL_D_END}"\n')

        return header + body

    def generate_did_files(self):
        """Generate required DID API files.
        Only use for some projects, which doesn't copy static code."""
        file_contents = {
            '.h': self.get_header_file_content(),
            '.c': self.get_source_file_content()
        }
        src_dst_dir = self._build_cfg.get_src_code_dst_dir()
        for extension, content in file_contents.items():
            file_path = Path(src_dst_dir, self.file_name + extension)
            with file_path.open(mode='w', encoding='utf-8') as file_handler:
                file_handler.writelines(content)


class ZCDIDs(ProblemLogger):
    """A class for handling of ZC DID definitions."""

    def __init__(self, build_cfg, unit_cfgs):
        """Init.

        Args:
            build_cfg (BuildProjConfig): Project configuration
            unit_cfgs (UnitConfigs): Unit definitions
        """
        super().__init__()
        self._build_cfg = build_cfg
        self._unit_cfgs = unit_cfgs
        self._valid_dids = None
        prefix = self._build_cfg.get_scheduler_prefix()
        self.operation_file_name = 'VcDIDAPI'
        self.sender_receiver_file_name = f'{prefix}UpdatingDIDValues'
        self.project_dids = self._get_project_dids()

    @property
    def valid_dids(self):
        return self._valid_dids

    @valid_dids.setter
    def valid_dids(self, yaml_dids):
        """Return a set of DIDs appearing in both the project and the project yaml file.

        Args:
            yaml_dids (dict): DIDs listed in the DID configuration yaml file.
        Returns:
            valid_dids (dict): Validated DIDs listed in both DID configuration yaml file as well as project.
        """
        self._valid_dids = {}

        dids_not_in_yaml = set(self.project_dids.keys()) - set(yaml_dids.keys())
        for did in dids_not_in_yaml:
            self.critical(f'DID {did} not defined in project diagnostics yaml file.')

        for did, did_data in yaml_dids.items():
            if did_data.get('manual', False):
                self._valid_dids[did] = {k: v for k, v in did_data.items() if k != "manual"}
                continue
            if did not in self.project_dids:
                self.warning(f'Ignoring DID {did}, not defined in any model.')
                continue
            data_type = self.project_dids[did]['type']
            if not data_type.startswith('UInt'):
                self.warning(f'Ignoring DID {did} of type {data_type}, only unsigned integers are supported.')
                continue
            self._valid_dids[did] = did_data

    def _get_project_dids(self):
        """Return a dict with DIDs defined in the project.
        Throws a critical error if something goes wrong.

        Returns:
            project_dids (dict): a dict with all dids in the project.
        """
        get_did_error_messages, project_dids = get_dids_in_prj(self._unit_cfgs)
        if get_did_error_messages:
            self.critical('\n'.join(get_did_error_messages))
            return {}
        return project_dids

    def _get_sender_receiver_header_file_content(self):
        """Get content for the S/R DID API header file.

        The function in this file is a runnable generated by yaml2arxml.

        Returns:
            (bool): True if any S/R DIDs are defined.
            (list(str)): List of lines to write to the S/R DID API header file.
        """
        name = self._build_cfg.get_composition_config("softwareComponentName")
        header_guard = f'{self.sender_receiver_file_name.upper()}_H'
        header = [
            f'#ifndef {header_guard}\n',
            f'#define {header_guard}\n',
            '\n',
            '#include "tl_basetypes.h"\n',
            f'#include "Rte_{name}.h"\n',
            '\n'
        ]
        footer = [f'\n#endif /* {header_guard} */\n']

        if not self.valid_dids:
            return False, header + footer

        variable_declarations = []
        function_declarations = []
        sender_receiver_dids_exist = False
        for did, did_data in self.valid_dids.items():
            project_did_data = self.project_dids[did]
            if did_data.get('PortType', 'dummy') not in ['BOTH', 'SENDER-RECEIVER']:
                continue  # C/R DIDs are handled in _get_operation... functions
            if not project_did_data['type'].startswith('UInt'):
                self.warning(
                    f'Ignoring DID {did} of type {project_did_data["type"]}, only unsigned integers are supported.'
                )
                continue
            sender_receiver_dids_exist = True
            define = self.project_dids[did]["class"].split('/')[-1]  # E.q. for ASIL D it is ASIL_D/CVC_DISP_ASIL_D
            variable_declarations.append(
                f'extern {define} {self.project_dids[did]["type"]} {self.project_dids[did]["name"]};\n'
            )
            function_declarations.append(
                f'extern UInt8 Rte_Write_DataServices_DID_{did}_data({project_did_data["type"]} {did});\n'
            )

        function_declarations.append(f'\nvoid Run_{self.sender_receiver_file_name}(void);\n')

        body = [
            f'#include "{build_defs.PREDECL_DISP_ASIL_D_START}"\n',
            *variable_declarations,
            f'#include "{build_defs.PREDECL_DISP_ASIL_D_END}"\n',
            f'\n#include "{build_defs.PREDECL_CODE_ASIL_D_START}"\n',
            *function_declarations,
            f'#include "{build_defs.PREDECL_CODE_ASIL_D_END}"\n',
        ]

        return sender_receiver_dids_exist, header + body + footer

    def _get_sender_receiver_source_file_content(self):
        """Get content for the S/R DID API source file.

        The function in this file is a runnable generated by yaml2arxml.

        Returns:
            (bool): True if any S/R DIDs are defined.
            (list(str)): List of lines to write to the S/R DID API source file.
        """
        header = [
            f'#include "{self.sender_receiver_file_name}.h"\n',
            '\n'
        ]

        if not self.valid_dids:
            return False, header

        sender_receiver_dids_exist = False
        body = [
            f'#include "{build_defs.CVC_CODE_ASIL_D_START}"\n',
            f'void Run_{self.sender_receiver_file_name}(void)\n',
            '{\n'
        ]
        for did, did_data in self.valid_dids.items():
            project_did_data = self.project_dids[did]
            if did_data.get('PortType', 'dummy') not in ['BOTH', 'SENDER-RECEIVER']:
                continue  # C/R DIDs are handled in _get_operation... functions
            if not project_did_data['type'].startswith('UInt'):
                self.warning(
                    f'Ignoring DID {did} of type {project_did_data["type"]}, only unsigned integers are supported.'
                )
                continue
            sender_receiver_dids_exist = True
            body.append(f'  Rte_Write_DataServices_DID_{did}_data({did});\n')
        body.extend([
            '}\n',
            f'#include "{build_defs.CVC_CODE_ASIL_D_END}"\n'
        ])

        return sender_receiver_dids_exist, header + body

    def _get_operation_data(self, operation, did_data):
        """Get operation function data of supported operations.

        Args:
            operation (str): Operation to get data for.
            did_data (dict): DID data.
        Returns:
            (dict): Operation function data.
        """
        array_size = byte_size(did_data['type'])
        if array_size > 1:
            read_data_declaration = f'UInt8 Run_{did_data["name"]}_ReadData(UInt8 Data[{array_size}])'
            read_data_definition = (
                '{\n'
                f'  for (UInt8 i = 0U; i < {array_size}; i++) {{\n'
                f'    Data[{array_size} - 1 - i] = ({did_data["name"]} >> (8 * i)) & 0xFF;\n'
                '  }\n'
                '  return 0U;\n'
                '}\n'
            )
        else:
            read_data_declaration = f'UInt8 Run_{did_data["name"]}_ReadData(UInt8 *Data)'
            read_data_definition = (
                '{\n'
                f'  *Data = {did_data["name"]};\n'
                '  return 0U;\n'
                '}\n'
            )
        operation_data = {
            'ReadData': {
                'declaration': read_data_declaration,
                'body': read_data_definition,
            }
        }
        if operation not in operation_data:
            return None
        return operation_data[operation]

    def _get_operation_header_file_content(self):
        """Get content for the DID API header file.

        Returns:
            (list(str)): List of lines to write to the DID API header file.
        """
        name = self._build_cfg.get_composition_config("softwareComponentName")
        header_guard = f'{self.operation_file_name.upper()}_H'
        header = [
            f'#ifndef {header_guard}\n',
            f'#define {header_guard}\n',
            '\n',
            '#include "tl_basetypes.h"\n',
            f'#include "Rte_{name}.h"\n',
            '\n'
        ]
        footer = [f'\n#endif /* {header_guard} */\n']

        if not self.valid_dids:
            return header + footer

        variable_declarations = []
        function_declarations = []
        for did, did_data in self.valid_dids.items():
            if did_data.get('PortType', 'dummy') == 'SENDER-RECEIVER':
                continue  # S/R DIDs are handled in _get_sender_receiver... functions
            define = self.project_dids[did]["class"].split('/')[-1]  # E.q. for ASIL D it is ASIL_D/CVC_DISP_ASIL_D
            variable_declarations.append(
                f'extern {define} {self.project_dids[did]["type"]} {self.project_dids[did]["name"]};\n'
            )
            for operation in did_data["operations"]:
                operation_data = self._get_operation_data(operation, self.project_dids[did])
                if operation_data is None:
                    self.warning(
                        f'Will not generate code for unsupported operation {operation}. Add manually for DID {did}.'
                    )
                    continue
                function_declarations.append(operation_data['declaration'] + ';\n')

        body = [
            f'#include "{build_defs.PREDECL_DISP_ASIL_D_START}"\n',
            *variable_declarations,
            f'#include "{build_defs.PREDECL_DISP_ASIL_D_END}"\n',
            f'\n#include "{build_defs.PREDECL_CODE_ASIL_D_START}"\n',
            *function_declarations,
            f'#include "{build_defs.PREDECL_CODE_ASIL_D_END}"\n',
        ]

        return header + body + footer

    def _get_operation_source_file_content(self):
        """Get content for the DID API source file.

        Returns:
            (list(str)): List of lines to write to the DID API source file.
        """
        header = [
            f'#include "{self.operation_file_name}.h"\n',
            '\n'
        ]

        if not self.valid_dids:
            return header

        body = [f'#include "{build_defs.CVC_CODE_ASIL_D_START}"\n']
        for did, did_data in self.valid_dids.items():
            if did_data.get('PortType', 'dummy') == 'SENDER-RECEIVER':
                continue  # S/R DIDs are handled in _get_sender_receiver... functions
            for operation in did_data["operations"]:
                operation_data = self._get_operation_data(operation, self.project_dids[did])
                if operation_data is None:
                    continue  # Warning already given in header generation
                body.append(operation_data['declaration'] + '\n' + operation_data['body'])
        body.append(f'#include "{build_defs.CVC_CODE_ASIL_D_END}"\n')

        return header + body

    def generate_did_files(self):
        """Generate required DID API files.
        Only use for some projects, which doesn't copy static code."""
        if self.valid_dids is None:
            includeDiagnostics = self._build_cfg.get_composition_config("includeDiagnostics")
            if includeDiagnostics in ["manual", "manual_dids"]:
                self.warning(f'includeDiagnostics is set to {includeDiagnostics}, not generating DID files.')
                return
            self.critical('Valid DIDs not set. Cannot generate DID files.')
            return

        src_dst_dir = self._build_cfg.get_src_code_dst_dir()

        # CLIENT-SERVER DIDs
        file_contents = {
            '.h': self._get_operation_header_file_content(),
            '.c': self._get_operation_source_file_content()
        }
        for extension, content in file_contents.items():
            file_path = Path(src_dst_dir, self.operation_file_name + extension)
            with file_path.open(mode='w', encoding='utf-8') as file_handler:
                file_handler.writelines(content)

        # SENDER-RECEIVER DIDs
        sender_receiver_dids_exist_header, header_contents = self._get_sender_receiver_header_file_content()
        sender_receiver_dids_exist_source, source_contents = self._get_sender_receiver_source_file_content()
        if sender_receiver_dids_exist_header and sender_receiver_dids_exist_source:
            file_path = Path(src_dst_dir, self.sender_receiver_file_name + '.h')
            with file_path.open(mode='w', encoding='utf-8') as file_handler:
                file_handler.writelines(header_contents)
            file_path = Path(src_dst_dir, self.sender_receiver_file_name + '.c')
            with file_path.open(mode='w', encoding='utf-8') as file_handler:
                file_handler.writelines(source_contents)



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module containing a problem logger class."""


class ProblemLogger:
    """Log helper class.

    Inherit from class ProblemLogger to log problems.

    Please note that all methods listed below are class methods
    meaning that all objects of classes inheriting ProblemLogger share a common logging stream.
    """

    __critical = []
    __warning = []
    __log = None

    @classmethod
    def init_logger(cls, logger):
        """Associates a system logger instance with ProblemLogger."""
        logger.info('Init logger.')
        cls.__log = logger

    @classmethod
    def clear_log(cls):
        """Clear internal error and warning log."""
        cls.__critical = []
        cls.__warning = []

    @classmethod
    def critical(cls, string, *args):
        """Log critical error message."""
        cls.__critical.append(string % args)
        if cls.__log is not None:
            cls.__log.critical(string, *args)

    @classmethod
    def warning(cls, string, *args):
        """Log warning message."""
        cls.__warning.append(string % args)
        if cls.__log is not None:
            cls.__log.warning(string, *args)

    @classmethod
    def info(cls, string, *args):
        """Log information message."""
        if cls.__log is not None:
            cls.__log.info(string, *args)

    @classmethod
    def debug(cls, string, *args):
        """Log debug message."""
        if cls.__log is not None:
            cls.__log.debug(string, *args)

    @classmethod
    def errors(cls):
        """Check if critical errors have been logged.

        Returns (boolean): True if critical errors have been logged.
        """
        return len(cls.__critical) > 0

    @classmethod
    def get_problems(cls):
        """Get logged problems.

        Returns (dict): a dict with error and warning messages.
        """
        return {
            'critical': cls.__critical,
            'warning': cls.__warning
        }

    @classmethod
    def get_nbr_problems(cls):
        """Get number of logged problems.

        Returns (dict): a dict with error and warning message counts.
        """
        return {
            'critical': len(cls.__critical),
            'warning': len(cls.__warning)
        }



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module containing classes for generation of signal consistency check report."""
from powertrain_build.signal_incons_html_rep_base import SigConsHtmlReportBase


class SigConsHtmlReport(SigConsHtmlReportBase):
    """Generate html report from the signal consistency check result.

    (see :doc:`signal_interfaces`)

    Inherits :doc:`SigConsHtmlReportBase <signal_incons_html_rep_base>`.
    """

    __intro = """<h2 class="nocount">Table of contents</h2>
    <ol>
      <li><a href="#introduction">Introduction</a></li>
      <li><a href="#unit_details">Detailed unit information</a></li>
      <li><a href="#ext_sigs">External signals</a>
      <ol>
        <li><a href="#ext_missing">Missing external signals</a></li>
        <li><a href="#ext_unused">Unused external signals</a></li>
      </ol></li>
      <li><a href="#unit_index">Unit index</a></li>
    </ol>
    <h2 id="introduction">Introduction</h2>
    <p>This documents lists inconsistencies in the internal and external signal configuration.</p>
    """

    def __init__(self, res_dict=None):
        """Initialize class instance.

        Args:
            res_dict (dict): result dict from the signal interface consistency check

        The dict shall have the following format::

            {
                "sigs": { "ext": {"missing": {},
                                  "unused": {},
                                  "inconsistent_defs": {}},
                          "int": {"UNIT_NAME": {"missing": {},
                                                "unused": {},
                                                "multiple_defs": {}
                                                "inconsistent_defs": {}}
                }
            }
        """
        super().__init__(res_dict)
        self.set_res_dict(res_dict)

    def _gen_unit_toc(self):
        """Generate a unit TOC for the unit with signal inconsistencies.

        Hyperlinks to more in depth unit information.
        """
        out = '  <h2 id="unit_index">Unit index</h2>\n'
        for unit in sorted(self._ext_units | self._int_units):
            out += f'  <div><a href="#{unit}">{unit}</a></div>\n'
        return out

    def _gen_units(self):
        """Generate all the information regarding all the units."""
        out = '  <h2 id="unit_details">Detailed Unit Information</h2>\n'
        for unit in sorted(self._ext_units | self._int_units):
            out += self._gen_unit(unit)
        return out

    def _gen_unit(self, unit):
        """Generate the html-report for the unit specific information."""
        out = f'  <h3 id="{unit}">{unit}</h3>\n'
        if unit in self._int_units:
            unit_data = self._res_dict['sigs']['int'][unit]
            out += self._gen_missing_sigs(unit_data)
            out += self._gen_unused_sigs(unit_data)
            out += self._gen_multiple_def_sigs(unit_data)
            out += self._gen_int_inconsistent_defs(unit_data)
        if unit in self._ext_units:
            out += self._gen_ext_inconsistent_defs(unit)
        return out

    def _gen_missing_sigs(self, unit_data, out=''):
        """Generate the unit specific information for missing signal in a unit."""
        out = '  <h4>Missing signals</h4>\n' \
              '  <p>Inports whose signals are not generated in the ' \
              'listed configuration(s).</p>'
        return super()._gen_missing_sigs(unit_data, out)

    def _gen_unused_sigs(self, unit_data, out=''):
        """Generate the unit specific information for the unused signals wihtin a unit."""
        out = '  <h4>Unused signals</h4>\n' \
              '  <p>Outports that are generated, but not used in the ' \
              'listed configuration(s).</p>'
        return super()._gen_unused_sigs(unit_data, out)

    def _gen_multiple_def_sigs(self, unit_data, out=''):
        """Generate unit specific information for the signals that are generated more than once."""
        out = '  <h4>Multiple defined signals</h4>\n' \
              '  <p>Outports that are generated more than once in ' \
              'the listed configuration(s).</p>'
        return super()._gen_multiple_def_sigs(unit_data, out)

    def _gen_ext_inconsistent_defs(self, unit, out=''):
        """Generate a report of inconsistent variable definition parameters.

        Report inconsistencies between the producing signal definition, and the
        signal definitions in the external interface definition.
        """
        out = '  <h4>External signal inconsistencies</h4>\n' \
              '  <p>In-/Out-ports that have different variable definitions ' \
              'than in the interface definition file.</p>'
        return super()._gen_ext_inconsistent_defs(unit, out)

    def _gen_int_inconsistent_defs(self, unit_data, out=''):
        """Generate a report of inconsistent variable definition parameters.

        Inconsistent for between the producing signal definition, and the
        consuming signal definitions for SPM internal signals.
        """
        out = '  <h4>Internal signal inconsistencies</h4>\n' \
              '  <p>Inports that have different variable definitions ' \
              'than the producing outport.</p>'
        return super()._gen_int_inconsistent_defs(unit_data, out)

    def _gen_ext_signals_report(self, type_, comment):
        """Generate report for external signals."""
        out = f'  <h3 id="ext_{type_}">{type_.capitalize()} external signals</h3>\n'
        out += f'<p>{comment}</p>'
        try:
            ext_data = self._res_dict['sigs']['ext'][type_]
            out += self._table_unused
            for var in sorted(ext_data.keys()):
                configs = ext_data[var]
                out += '  <tr>\n'
                out += f'        <td>{var}</td>\n'
                out += f'        <td>{self._set_to_str(configs)}</td>\n'
                out += '      </tr>\n'
            out += '    </tbody>\n'
            out += '  </table>\n'
        except KeyError:
            pass
        return out

    def set_res_dict(self, res_dict):
        """Set the result dictionary used to generate the html-report.

        Args:
            res_dict (dict): result dict from the signal interface consistency check

        See class constructor for dict structure.
        """
        self._res_dict = res_dict
        self._ext_units = set()
        self._int_units = set()
        if res_dict is not None and 'sigs' in res_dict:
            if 'ext' in res_dict['sigs'] and 'inconsistent_defs' in res_dict['sigs']['ext']:
                self._ext_units = set(self._res_dict['sigs']['ext']['inconsistent_defs'].keys())
            if 'int' in res_dict['sigs']:
                self._int_units = set(self._res_dict['sigs']['int'].keys())

    def gen_contents(self):
        """Generate report contents from the signal inconsistency check result dictionary.

        Overrides HtmlReport.gen_contents()
        """
        html = []
        html += self.__intro
        html += self._gen_units()
        html += '  <h2 id="ext_sigs">Signals missing and unused in the ' \
                'interface definition</h2>\n'
        html += self._gen_ext_signals_report('missing', 'Signals not generated by '
                                                        'Vcc SW, but are defined in '
                                                        'the Interface definition to be '
                                                        'generated')
        html += self._gen_ext_signals_report('unused', 'Signals defined to be generated by '
                                                       'supplier SW, but are not used '
                                                       'by VCC SW')
        html += self._gen_unit_toc()
        return ''.join(html)



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module for CSP API abstraction."""

import enum
from ruamel.yaml import YAML
from abc import abstractmethod
from powertrain_build.interface.base import BaseApplication, Signal
from powertrain_build.lib import logger

LOGGER = logger.create_logger("service")


class MissingApi(Exception):
    """Exception to raise when api is missing"""
    def __init__(self, api, map_file):
        self.message = f"Api {api} missing from {map_file}"


class BadYamlFormat(Exception):
    """Exception to raise when in/out signal is not defined."""
    def __init__(self, api, signal_name):
        self.message = f"Signal {signal_name} for {api} should be set as insignal or outsignal."


class CspApi(BaseApplication):
    """Abstraction for HAL and SFW"""
    position = enum.Enum(
        'Position',
        names=[
            "property_name",
            "property_type",
            "variable_type",
            "offset",
            "factor",
            "default",
            "length",
            "min",
            "max",
            "enum",
            "init",
            "description",
            "unit",
            "endpoint",
            "api",
            "variant",
            "strategy",
            "debug",
            "dependability"
        ]
    )

    def __init__(self, base_application, read_strategy='Always'):
        """Create the interface object

        Args:
            base_application (BaseApplication): Primary object of an interface
                                                Usually a raster, but can be an application or a model too.
        """
        self.name = ""
        self.translations = {}
        self.api_side = False
        # we only care when generating models for csp.
        self.filter = None
        self.signal_names = {
            "api": {"insignals": set(), "outsignals": set()},
            "app": {"insignals": set(), "outsignals": set()},
        }
        self.base_application = base_application
        self.map_file = self.get_map_file()
        self.api_map = self.get_map()
        self.translations_files = []
        self.signal_primitives_list = []
        self.default_read_strategy = read_strategy

    def get_signal_properties(self, signal):
        """Get signal properties for signal

        Calls self.base_application to get signal properties

        Args:
            signal (Signal): Signal to get properties for
        """
        self.base_application.get_signal_properties(signal)

    def _get_signals(self):
        """Read signals"""
        self.parse_definition(self.translations_files)

    @property
    def insignals(self):
        return self.get_signals(self.hal_name, 'insignals')

    @property
    def outsignals(self):
        return self.get_signals(self.hal_name, 'outsignals')

    def get_signals(self, api, signal_type='insignals'):
        """Get signals to and from an api abstraction

        self.api_side configures if we look at the api side.
        If it is set to False, we look at the application side.

        Args:
            api (str): Name of the api
            signal_type (str): insignals or outsignals
        Returns:
            signals (list): Signals in the interface
        """
        if self.api_side:
            signal_names = self.signal_names['api'][signal_type]
        else:
            signal_names = self.signal_names['app'][signal_type]

        signals = []
        for signal_name, specs in self.translations.items():
            for spec in specs:
                signal = None
                if api is not None:
                    if spec['api'] == api:
                        if self.api_side:
                            if spec['property'] in signal_names:
                                signal = Signal(spec['property'], self)
                        else:
                            if signal_name in signal_names:
                                signal = Signal(signal_name, self)
                else:
                    if self.api_side:
                        if spec['property'] in signal_names:
                            signal = Signal(spec['property'], self)
                    else:
                        if signal_name in signal_names:
                            signal = Signal(signal_name, self)
                if signal is not None:
                    signals.append(signal)
        return signals

    def clear_signal_names(self):
        """Clear signal names

        Clears defined signal names (but not signal properties).
        """
        self.signal_names = {
            "api": {"insignals": set(), "outsignals": set()},
            "app": {"insignals": set(), "outsignals": set()},
        }

    def parse_definition(self, definition):
        """Parses all definition files

        Args:
            definition (list(Path)): Definition files
        """
        for translation in definition:
            raw = self.read_translation(translation)
            self.extract_endpoint_definitions(raw)

    @staticmethod
    def get_api_name(api_name):
        """Return the api name

        Args:
            api_name (str): Name of the api

        Returns:
            (str): Name of the api
        """
        return api_name

    def verify_api(self, api_name):
        """Verify that the api is in the map

        Args:
            api_name (str): Name of the api
        """
        if api_name not in self.api_map:
            raise MissingApi(api_name, self.map_file)

    def add_signals(self, signals, signal_type='insignals', properties=None):
        """Add signal names and properties to already set ones

        Args:
            signals (list(Signals)): Signals to use
            signal_type (str): 'insignals' or 'outsignals'
            properties (list(str)): signal definition properties, default = []
        """
        opposite = {'insignals': 'outsignals', 'outsignals': 'insignals'}
        api_type = opposite[signal_type]
        properties = [] if properties is None else properties
        for signal in signals:
            LOGGER.debug("Adding signal: %s", signal)
            temp_set = set()
            for translation in self.translations.get(signal.name, []):
                temp_list = list(translation)
                api_name = translation[self.position.api.value]
                variant_name = translation[self.position.variant.value]
                endpoint = translation[self.position.endpoint.value]
                api_signal = translation[self.position.property_name.value]
                self.check_signal_property(api_name, variant_name, endpoint,
                                           api_signal, signal_type)
                self.signal_names['api'][api_type].add(api_signal)
                for enum_property in properties:
                    LOGGER.debug("Modifying property: %s", enum_property)
                    value = signal.properties[enum_property["source"]]
                    if value == "-":
                        value = enum_property["default"]
                    temp_list[
                        self.position[enum_property["destination"]].value
                    ] = value
                temp_set.add(tuple(temp_list))
                self.translations[signal.name] = temp_set
            self.signal_names['app'][signal_type].add(signal.name)
        self.check_endpoints()
        LOGGER.debug('Registered signal names: %s', self.signal_names)

    def check_signal_property(self, api, variant, endpoint, property_name, signal_type):
        """Check if we have only one signal written for the same property.

        Args:
            api (str): interface name
            variant (str): variant value. "properties" or "methods" for service
                           "hals" for hal
            endpoint (str): signal endpoint
            property_name (str): signal property
            signal_type (str): 'insignals' or 'outsignals'
        """
        primitive_value = ""
        for value in [api, variant, endpoint, property_name]:
            if value:
                if primitive_value == "":
                    primitive_value = value
                else:
                    primitive_value = primitive_value + '.' + value
        if primitive_value == "":
            raise Exception("The primitive does not contain any value!")
        directional_primitive = f"{primitive_value}.{signal_type}"
        self.check_property(directional_primitive, signal_type)

    def check_property(self, property_spec, signal_type):
        """Check if we have only one signal written for the same property.

        Args:
            property_spec (str): property specification
            signal_type (str): 'insignals' or 'outsignals'
        """
        if property_spec in self.signal_primitives_list:
            error_msg = (f"You can't write {property_spec} as "
                         f"{signal_type} since this primitive has been used."
                         " Run model_yaml_verification to identify exact models.")
            raise Exception(error_msg)
        self.signal_primitives_list.append(property_spec)

    @abstractmethod
    def check_endpoints(self):
        """Should be implemented by subclasses."""

    @staticmethod
    def read_translation(translation_file):
        """Read specification of the format:

        service:
          interface:
            properties:
              - endpoint_name:
                - signal: name
                  property: name
                - signal: name
                  property: name
        hal:
          hal_name:
            - primitive_endpoint:
              - insignal: name
          hal_name:
            - struct_endpoint:
              - insignal: name1
                property: member1
              - insignal: name2
                property: member2
          ecm:
            - signal: name
        signals:
          tvrl:
            - signal: name
              property: can_name
              offset: offset
              factor: scaling

        Args:
            translation_file (Path): file with specs

        Returns:
            yaml_data (dict): Loaded YAML data as dict, empty if not found
        """
        if not translation_file.is_file():
            LOGGER.warning("No file found for %s", translation_file)
            return {}
        with open(translation_file, encoding="utf-8") as translation:
            yaml = YAML(typ='safe', pure=True)
            raw = yaml.load(translation)
        return raw

    def parse_api_definitions(self, api_definitions):
        """Parses group definitions.

        Args:
            api_definitions (dict): endpoints in parsed yaml file.
        """
        for api_from_spec, definition in api_definitions.items():
            for variant, variant_endpoints in self.extract_definition(definition).items():
                for endpoints in variant_endpoints:
                    for endpoint, signals in endpoints.items():
                        self.parse_property_definitions({api_from_spec: signals}, endpoint, variant)

    def parse_property_definitions(self, endpoint_definitions, endpoint, variant):
        """Parse signal definitions.

        Args:
            endpoint_definitions (dict): parsed yaml file.
            endpoint (str): Name of the endpoint to use
        """
        def _get_property_name(specification):
            """ Handle cases when there are no propery name.

            If there is no property name, the "group" is set to the signal.
            This should be used when the property is not a struct in the interface api.

            Args:
                specification (dict): signal specification
            Returns:
                property_name (str): name of the potential internal property
            """
            property_name = specification.get('property', '')
            if not property_name:
                return None
            return property_name

        enumerations = self.base_application.enumerations

        for api, specifications in endpoint_definitions.items():
            self.verify_api(api)
            for specification in specifications:
                in_out_signal = [key for key in specification.keys() if 'signal' in key]
                base_signal = None
                signal_name = None
                if "in" in in_out_signal[0]:
                    for signal in self.base_application.insignals:
                        if signal.name == specification["insignal"]:
                            base_signal = signal
                            signal_name = signal.name
                elif "out" in in_out_signal[0]:
                    for signal in self.base_application.outsignals:
                        if signal.name == specification["outsignal"]:
                            base_signal = signal
                            signal_name = signal.name
                else:
                    raise BadYamlFormat(api, specification[in_out_signal[0]])
                if base_signal is None:
                    continue
                base_properties = self.base_application.get_signal_properties(
                    base_signal
                )

                if base_properties["type"] in enumerations:
                    underlying_data_type = enumerations[base_properties['type']]['underlying_data_type']
                    interface_type = underlying_data_type
                    if 'init' not in specification:
                        if enumerations[base_properties['type']]['default_value'] is not None:
                            init_value = enumerations[base_properties['type']]['default_value']
                        else:
                            LOGGER.warning('Initializing enumeration %s to "zero".', base_properties['type'])
                            init_value = [
                                k for k, v in enumerations[base_properties['type']]['members'].items() if v == 0
                            ][0]
                    else:
                        init_value = specification.get("init", 0)
                else:
                    interface_type = base_properties["type"]
                    init_value = specification.get("init", 0)

                if "out" in in_out_signal[0] and "strategy" in specification:
                    LOGGER.warning('Cannot set read strategy for outsignal %s, using "Always".', signal_name)
                    strategy = "Always"
                else:
                    strategy = specification.get("strategy", self.default_read_strategy)
                if strategy not in self.read_strategies:
                    LOGGER.warning('Invalid strategy %s, using "Always" instead.', strategy)
                    strategy = self.default_read_strategy

                if signal_name not in self.translations:
                    self.translations[signal_name] = set()
                self.translations[signal_name].add(
                    (
                        "enum_0",  # Enum starts at position 1.
                        _get_property_name(specification),
                        interface_type,
                        specification.get("type"),
                        specification.get("offset"),
                        specification.get("factor"),
                        specification.get("default"),
                        specification.get("length"),
                        specification.get("min"),
                        specification.get("max"),
                        specification.get("enum"),
                        init_value,
                        specification.get("description"),
                        specification.get("unit"),
                        endpoint,
                        api,
                        variant,
                        strategy,
                        specification.get("debug", False),
                        specification.get("dependability", False)
                    )
                )

    def spec_to_dict(self, signal_spec, signal_name):
        """Convert signal specification to dict

        Args:
            signal_spec (dict): signal specification
            signal_name (str): signal name
        Returns:
            (dict): signal specification as dict
        """
        return {
            'variable': signal_name,
            'variable_type': signal_spec[self.position.variable_type.value],
            'property': signal_spec[self.position.property_name.value],
            'property_type': signal_spec[self.position.property_type.value],
            "default": signal_spec[self.position.default.value],
            "length": signal_spec[self.position.length.value],
            'offset': signal_spec[self.position.offset.value],
            'factor': signal_spec[self.position.factor.value],
            'range': {
                'min': signal_spec[self.position.min.value],
                'max': signal_spec[self.position.max.value]
            },
            'init': signal_spec[self.position.init.value],
            'description': signal_spec[self.position.description.value],
            'unit': signal_spec[self.position.unit.value],
            'endpoint': signal_spec[self.position.endpoint.value],
            'api': self.get_api_name(signal_spec[self.position.api.value]),
            'variant': signal_spec[self.position.variant.value],
            'strategy': signal_spec[self.position.strategy.value],
            'debug': signal_spec[self.position.debug.value],
            'dependability': signal_spec[self.position.dependability.value]
        }

    def to_dict(self, client="app"):
        """Method to generate dict to be saved as yaml

        Returns:
            spec (dict): Signalling specification
        """
        spec = {"consumer": [], "producer": []}
        direction = {
            "app": ["consumer", "producer"],
            "api": ["producer", "consumer"]}
        for signal_name, signal_spec in self._generator(self.signal_names["app"]["insignals"]):
            spec[direction[client][0]].append(
                self.spec_to_dict(signal_spec, signal_name)
            )
        for signal_name, signal_spec in self._generator(self.signal_names["app"]["outsignals"]):
            spec[direction[client][1]].append(
                self.spec_to_dict(signal_spec, signal_name)
            )
        return spec

    def _generator(self, signal_names, unique_names=False):
        """Iterate over signals for allowed services

        If unique_names is True, the iterator does not yield the same signal twice
        if unique_names is False, it yields each allowed signal spec with the signal name

        Args:
            signal_names (list): allowed signals
        Yields:
            name (str): Name of the signal
            specification (dict): signal specification for allowed service
        """
        for signal_name, specifications in (
                (name, spec) for name, spec in sorted(
                    self.translations.items())
                if name in signal_names):
            for specification in specifications:
                if unique_names:
                    yield signal_name, specification
                    break
                yield signal_name, specification



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

""" script for running running signal consistency check on specific models. Git HEAD or local.
"""

import argparse
import csv
import os
import sys
from collections import defaultdict
from os.path import join
from pathlib import Path
from typing import List, Optional

import git

from powertrain_build.build_proj_config import BuildProjConfig
from powertrain_build.feature_configs import FeatureConfigs
from powertrain_build.lib import helper_functions, logger
from powertrain_build.signal_incons_html_rep_all import SigConsHtmlReportAll
from powertrain_build.signal_interfaces import CsvSignalInterfaces, YamlSignalInterfaces
from powertrain_build.unit_configs import UnitConfigs
from powertrain_build.user_defined_types import UserDefinedTypes

LOGGER = logger.create_logger(__file__)
EXIT_CODE_OK = 0
EXIT_CODE_UNPRODUCED = 1
EXIT_CODE_MISSING_CONSUMER = 2
# Files define in <mdl>.Unconsumed.csv is consumed by mdl or/and interface.
EXIT_CODE_INCORRECT_CSV = 4
EXIT_CODE_NEVER_ACTIVE_SIGNALS = 8
UNCONSUMED_SIGNAL_FILE_PATTERN = "*_Unconsumed_Sig.csv"
MODEL_ROOT_DIR = "Models"
REPO_ROOT = helper_functions.get_repo_root()
MISSING_CONSUMER_CONSOLE_MESSAGE = (
    "\n===============Unused signals===============\n{signals}"
)
UNPRODUCED_SIGNALS_CONSOLE_MESSAGE = (
    "\n=============Unproduced signals=============\n{signals}"
)
UNCONSUMED_CSV_CONSOLE_MESSAGE = "\n=========Signals defined in unconsumed.csv-s that are consumed=========\n{signals}"
NEVER_ACTIVE_SIGNALS_CONSOLE_MESSAGE = (
    "\n============Never active signals=============\n{signals}"
)
INDEX_FILE = "Reports/Index_SigCheck_All.html"
TEMPLATE = """<!DOCTYPE html>
<!--[if IE 8]><html class="no-js lt-ie9" lang="en" > <![endif]-->
<!--[if gt IE 8]><!--> <html class="no-js" lang="en" > <!--<![endif]-->

<body class="wy-body-for-nav">

<a class="icon icon-home">      Signal Consistency Report
</a>
<ul>
<li class="toctree-l1">
    <a class="reference internal" href="SigCheckAll_intro.html">Introduction</a>
</li>
{project_rows}
</ul>
</body>
</html>"""

PARSER_HELP = "Run signal inconsistency check."


def gen_sig_incons_index_file(project_list):
    """Generate Index_SigCheck_All.html."""
    # TODO Remove function and add as method in powertrain_build/signal_incons_html_rep_all.py
    rows = ""
    project_row = """<li class="toctree-l1"><a class="reference internal"
                   href="SigCheckAll_{project}.html">SigCheckAll {project}</a></li>\n"""
    for prj in project_list:
        rows += project_row.format(project=prj)
    os.makedirs(os.path.dirname(INDEX_FILE), exist_ok=True)
    with open(INDEX_FILE, "w", encoding="utf-8") as f_h:
        f_h.write(TEMPLATE.format(project_rows=rows))


def configure_parser(parser: argparse.ArgumentParser):
    """Parse the arguments sent to the script."""
    parser.add_argument(
        '-p',
        '--project-folders',
        nargs='+',
        default=['Projects'],
        help='Directories to look for project configurations.'
    )
    parser.add_argument(
        "-m",
        "--models",
        nargs="+",
        help="Arguments for running in local repo. "
        "Name of models to run test on Ex: VcDepExt VcAcCtrl. "
        "If not supplied, changed files in current commit will be used. "
        "Reports are stored in local repo",
    )
    parser.add_argument(
        "-r", "--report", help="Create report for all projects", action="store_true"
    )
    parser.set_defaults(func=run_signal_inconsistency_check)


def get_project_configs(project_folders):
    """Wrapper for creating project configs for all projects.

    Args:
        project_folders (list): List of project folders to look for project configs in.
    Returns
        prj_cfgs (dict): Project configs.
    """
    project_config_files = []
    root_path = Path().resolve()
    for project_folder in project_folders:
        found_files = root_path.glob(f"{project_folder}/**/ProjectCfg.json")
        project_config_files.extend(list(map(lambda x: x.as_posix(), found_files)))
    prj_cfgs = {}
    for project_config_file in project_config_files:
        LOGGER.debug("Get project config for %s", project_config_file)
        project_config = BuildProjConfig(os.path.normpath(project_config_file))
        prj_cfgs.update({project_config.name: project_config})
    return prj_cfgs


def check_inconsistency(signal_ifs, never_active_signals):
    """Check signal inconsistency.
    Args:
        signal_ifs (CsvSignalInterfaces): class holding signal interface information.
        never_active_signals (dict): Dict of projects mapped to units mapped to "NEVER_ACTIVE" signals.
    Returns:
        inconsistency result (dict)
    """
    signal_inconsistency_results = {}
    for key, signal_if in signal_ifs.items():
        partial_result = signal_if.check_config()
        signal_inconsistency_results.update({key: partial_result})
        signal_inconsistency_results[key].update(
            {"never_active_signals": never_active_signals[key]}
        )
    return signal_inconsistency_results


def get_signal_interfaces(prj_cfgs, models):
    """Wrapper for creating signal interface dict for all configs.
    Args:
        Project config.
    Returns:
           Signal interface dict.
    """
    signal_ifs = {}
    per_unit_cfgs = {}
    for prj, prj_cfg in prj_cfgs.items():
        LOGGER.info("Parsing interface %s", prj)

        feature_cfg = FeatureConfigs(prj_cfg)
        unit_cfg = UnitConfigs(prj_cfg, feature_cfg)
        user_defined_types = UserDefinedTypes(prj_cfg, unit_cfg)

        if prj_cfg.has_yaml_interface:
            signal_if = YamlSignalInterfaces(
                prj_cfg, unit_cfg, feature_cfg, user_defined_types, model_names=models
            )
        else:
            signal_if = CsvSignalInterfaces(prj_cfg, unit_cfg, models)

        signal_ifs.update({prj: signal_if})
        per_unit_cfgs.update({prj: unit_cfg.get_per_unit_cfg()})

    return signal_ifs, per_unit_cfgs


def get_never_active_signals(prj_cfgs, models):
    """Functions for getting all signals marked as "NEVER_ACTIVE" in all projects.

    Args:
        prj_cfgs (list(BuildProjConfig)): List of project configurations to look for never active signals in.
        models (list): List of Simulink models to check.
    Returns:
        never_active_signals (dict): Dict of projects mapped to units mapped to its "NEVER_ACTIVE" signals.
    """
    never_active_signals = {}
    for prj, prj_cfg in prj_cfgs.items():
        LOGGER.info("searching for never active signals in %s", prj)
        feature_cfg = FeatureConfigs(prj_cfg)
        unit_cfg = UnitConfigs(prj_cfg, feature_cfg)
        per_unit_cfg_total = unit_cfg.get_per_unit_cfg_total()
        never_active_signals[prj] = {}
        for unit, unit_data in per_unit_cfg_total.items():
            if unit not in models:
                continue
            inactive_inports = []
            inactive_outports = []
            for inport, inport_data in unit_data["inports"].items():
                if "NEVER_ACTIVE" in inport_data["configs"]:
                    inactive_inports.append(inport)
            for outport, outport_data in unit_data["outports"].items():
                if "NEVER_ACTIVE" in outport_data["configs"]:
                    inactive_outports.append(outport)
            if inactive_inports or inactive_outports:
                never_active_signals[prj].update(
                    {unit: inactive_inports + inactive_outports}
                )

    return never_active_signals


class SignalInconsistency:
    """Class for running signal consistency check on specific models. Git HEAD or local."""

    def __init__(self, args):
        """Models in local repo."""
        self.never_active_signals = {}
        self.signal_inconsistency_results = {}
        self.per_unit_cfgs = {}
        self.signal_ifs = {}
        self.repo = git.Repo()
        if args.models:
            self.models = args.models
        else:
            mdl_file_paths = self.get_changed_models()
            self.models = self._get_model_names(mdl_file_paths)

    @staticmethod
    def _get_model_names(mdl_paths):
        """Get Change Model names."""
        mdl_index = 1
        return [os.path.split(i)[mdl_index].replace(".mdl", "") for i in mdl_paths]

    def _create_reports(self):
        """Create reports for all projects."""
        sig_report = SigConsHtmlReportAll(self.signal_inconsistency_results)
        sig_report.generate_report_file_signal_check(
            join(REPO_ROOT, "Reports", "SigCheckAll")
        )

    def _sort_internal_inconsistency_by_model(self):
        """Sort inconsistency by model."""
        sig = self.signal_inconsistency_results
        no_producer_int = {}
        missing_consumer_int = {}
        never_active_int = {}
        for project, val in sig.items():
            int_dict = val["sigs"]["int"]
            for mdl, data in int_dict.items():
                if data.get("missing"):
                    if mdl not in no_producer_int:
                        no_producer_int[mdl] = self._to_tuple_list(
                            project, data["missing"]
                        )
                    else:
                        no_producer_int[mdl] += self._to_tuple_list(
                            project, data["missing"]
                        )
                if data.get("unused"):
                    if mdl not in missing_consumer_int:
                        missing_consumer_int[mdl] = self._to_tuple_list(
                            project, data["unused"]
                        )
                    else:
                        missing_consumer_int[mdl] += self._to_tuple_list(
                            project, data["unused"]
                        )
            for mdl, never_active_signals in val["never_active_signals"].items():
                never_active_int[mdl] = [
                    (signal, {project}) for signal in never_active_signals
                ]

        return (
            self._merge_tuple_list(no_producer_int),
            self._merge_tuple_list(missing_consumer_int),
            self._merge_tuple_list(never_active_int),
        )

    def _analyse_inconsistency(self):
        """Collect signals.

        Args:
            no_producer_int (dict): Consumed signals but not produced.
            missing_consumer_int (dict): Produced signals but not consumed.
        """
        exit_code = EXIT_CODE_OK
        log_message = ""
        (
            no_producer_int,
            missing_consumer_int,
            never_active_int,
        ) = self._sort_internal_inconsistency_by_model()
        used_model_inports = self.aggregate_model_inports()
        used_external_outports = self.aggregate_supplier_inports()
        exit_code_unproduced, unproduced_message = self.check_unproduced_signals(
            no_producer_int
        )
        (
            exit_code_missing_consumer,
            missing_consumer_message,
        ) = self.check_missing_consumer_signals(missing_consumer_int)
        exit_code_used_inports, used_inports_message = self.check_unconsumed_files(
            used_model_inports, used_external_outports
        )
        exit_code_never_active, never_active_message = self.check_never_active_signals(
            never_active_int
        )
        log_message += unproduced_message
        log_message += missing_consumer_message
        log_message += used_inports_message
        log_message += never_active_message
        exit_code += exit_code_unproduced
        exit_code += exit_code_missing_consumer
        exit_code += exit_code_used_inports
        exit_code += exit_code_never_active
        return exit_code, log_message

    def get_changed_models(self):
        """Get changed models in current commit."""
        changed_files_tmp = self.repo.git.diff(
            "--diff-filter=d", "--name-only", "HEAD~1"
        )
        changed_files = changed_files_tmp.splitlines()
        changed_models = [
            m for m in changed_files if m.endswith(".mdl") or m.endswith(".slx")
        ]
        return changed_models

    def check_unproduced_signals(self, unproduced_signals):
        """Check if models expects signals that is not produced.
        Args:
            unproduced_signals (dict):
                                     mdl:{signals:{projects...}...}...
        Returns:
               exit_code (int):
                              Number of unproduced signals in checked models.
               console message (string):
                                      Shows if models has unproduced signals.
        """
        error_message = ""
        exit_code = EXIT_CODE_OK
        models_with_unproduced_signals = {
            m for m in unproduced_signals.keys() if m in self.models
        }
        for mdl in models_with_unproduced_signals:
            error_message += self._format_console_output(mdl, unproduced_signals[mdl])
            if mdl in unproduced_signals and unproduced_signals[mdl]:
                exit_code = EXIT_CODE_UNPRODUCED
        return exit_code, UNPRODUCED_SIGNALS_CONSOLE_MESSAGE.format(
            signals=error_message
        )

    def check_missing_consumer_signals(self, missing_consumer_signals):
        """Check if models expect signals that is not produced.
        Args:
            missing_consumer_signals (dict):
                                     mdl:{signals:{projects...}...}...
        Returns:
               exit_code (int):
                              Number of unproduced signals in checked models.
               console message (string):
                                      Shows if models has unproduced signals.
        """
        error_message = ""
        exit_code = EXIT_CODE_OK
        unconsumed_skip_list = self.fetch_signals_to_skip()
        models_producing_not_consumed_signals = {
            m for m in missing_consumer_signals.keys() if m in self.models
        }
        for mdl in models_producing_not_consumed_signals:
            sigs = missing_consumer_signals[mdl]
            for k in set(sigs.keys()):
                if mdl in unconsumed_skip_list and k in unconsumed_skip_list[mdl]:
                    del sigs[k]
            error_message += self._format_console_output(
                mdl, missing_consumer_signals[mdl]
            )
            if mdl in missing_consumer_signals and missing_consumer_signals[mdl]:
                exit_code = EXIT_CODE_MISSING_CONSUMER
        return exit_code, MISSING_CONSUMER_CONSOLE_MESSAGE.format(signals=error_message)

    def check_unconsumed_files(self, inports, ifs_outports):
        """Check that signals define in unconsumed.csv files are not being consumed
        Args:
            inports (set): all models inports in all projects.
            ifs_outports (set): all outports, all projects and all interfaces.
        Returns:
               exit_code (int):
               console message (string): Console error message (if test fail)
        """
        error_message_int = ""
        error_message_ext = ""
        error_message = ""
        exit_code = EXIT_CODE_OK
        unconsumed_skip_list = self.fetch_signals_to_skip()

        for producing_mdl, signals in unconsumed_skip_list.items():
            signal_intersect_int = signals & inports
            signal_intersect_ext = signals & ifs_outports
            if signal_intersect_int:
                exit_code, error_message_int = self.get_intersect_exit_code(
                    producing_mdl, signal_intersect_int
                )
            if signal_intersect_ext:
                exit_code, error_message_int = self.get_intersect_exit_code(
                    producing_mdl, signal_intersect_ext, False
                )
        error_message += error_message_int
        error_message += error_message_ext
        return exit_code, UNCONSUMED_CSV_CONSOLE_MESSAGE.format(signals=error_message)

    def check_never_active_signals(self, never_active_signals):
        """Check if there are any never active signals.
        Produce corresponding error code and message.

        Args:
            never_active_signals (dict): Dict mapping models to never active signals in projects.
        Returns:
            exit_code (int):
            console message (string): Console error message (if test fail)
        """
        error_message = ""
        exit_code = EXIT_CODE_OK
        if never_active_signals:
            exit_code = EXIT_CODE_NEVER_ACTIVE_SIGNALS
            for mdl, signals in never_active_signals.items():
                error_message += self._format_console_output(mdl, signals)

        return exit_code, NEVER_ACTIVE_SIGNALS_CONSOLE_MESSAGE.format(
            signals=error_message
        )

    def get_intersect_exit_code(
        self, producing_mdl, signal_intersect, int_signals=True
    ):
        """Determine if Unconsumed.csv and used inport signals overlap per project.
        Return exit code and which consumer and project is cause of failure.
        Args:
            producing_mdl (String): Model name.
            signal_intersect (set): Unconsumed.csv signals and used inports that overlap.
            int_signals (bool):
                              True: Set exit code for internally overlaping signals (Models)
                              False: Set exit code for external overlaping signals (Supplier)
        Returns:
               exit_code (int): Fail or succes
               error_message (String): Consumers.
        """
        exit_code = EXIT_CODE_OK
        console_message = "{producer}\n {signal} is consumed by {consumer} in {prj}\n"
        error_message = ""
        for sig in signal_intersect:
            if int_signals:
                consumers = self.get_consumer_int(sig)
            else:
                consumers = self.get_consumer_ext(sig)
            for unit, prj in consumers:
                if self.mdl_is_producer_in_prj(producing_mdl, prj, sig):
                    error_message += console_message.format(
                        producer=producing_mdl, signal=sig, consumer=unit, prj=prj
                    )
                    exit_code = EXIT_CODE_INCORRECT_CSV
        return exit_code, error_message

    def mdl_is_producer_in_prj(self, mdl, prj, signal):
        """Check that defined signal in <mdl>unconsumed.csv is produced in project"""
        try:
            return signal in self.per_unit_cfgs[prj][mdl]["outports"].keys()
        except KeyError:
            return False

    def get_consumer_int(self, signal):
        """Find internal consumers"""
        signal_consumers = []
        for prj, unit_cfgs in self.per_unit_cfgs.items():
            for unit, unit_cfg in unit_cfgs.items():
                if signal in unit_cfg["inports"].keys():
                    signal_consumers.append((unit, prj))
        return signal_consumers

    def get_consumer_ext(self, signal):
        """Find external consumers"""
        signal_consumers = []
        for prj, if_output in self.signal_ifs.items():
            for sig_type, signals in if_output.get_external_outputs(prj).items():
                if signal in signals.keys():
                    signal_consumers.append((sig_type, prj))
        return signal_consumers

    @staticmethod
    def _format_console_output(mdl, signal_dict):
        out = ""
        for signal, projects in signal_dict.items():
            out += " " + str(signal) + str(projects) + "\n"
        return f"\n{mdl}\n{out} \n"

    @staticmethod
    def _to_tuple_list(project, signal_project_dict):
        """Create list of tuples.

        NOTE: signal_project_dict is no longer a signal to project mapping.
        It was when PyBuild handled multiple projects at once.

        Args:
            project (str): project name.
            signal_project_dict (dict): signal1: {}, signal2: {}...
        Returns:
               list: [(signal1, {project}), (signal2, {project})...]
        """
        return [(signal, {project}) for signal in signal_project_dict.keys()]

    @staticmethod
    def _merge_tuple_list(t_list):
        """Merged list of tuples.
        Args (list(tuple...)):
                             [(A,1), (A,2), (C,1)]

        Returns (list(tuple...)):
                                [(A,[1,2]), (C,1)]
        """
        model_confs = {}
        for mdl, sig_prj_pair in t_list.items():
            model_conf = defaultdict(list)
            for signal, prj in sig_prj_pair:
                model_conf[signal].append(prj)
            model_confs[mdl] = model_conf
        return model_confs

    def fetch_signals_to_skip(self, model_root_dir=MODEL_ROOT_DIR):
        """fetch Unconsumed signals from csv files.
        Returns:
               unconsumed_signals (dict):
                                        {model: {sig1, sig2...}, model2: {sig1, sig2...}}
        """
        unconsumed_sig = {mdl: set() for mdl in self.models}
        # "mdl_name" = "mdl folder name". '..\\..\\mdl_name\\pybuild_cfg\\mdl_Unconsumed_Sig.csv'
        model_dir_index = -3
        signal_col_index = 0
        for file_name in Path(model_root_dir).glob(
            "**/" + UNCONSUMED_SIGNAL_FILE_PATTERN
        ):
            mdl_name = str(file_name).split(os.sep)[model_dir_index]
            if mdl_name in self.models:
                with file_name.open() as csv_file:
                    csv_reader = csv.reader(csv_file)
                    next(csv_reader)  # Skip header.
                    for row in csv_reader:
                        unconsumed_sig[mdl_name].add(row[signal_col_index])
        return unconsumed_sig

    @staticmethod
    def _print_console_log_info(exit_code):
        """Prints explanatory message to console.
        NOTE: The exit_codes_messages variable needs the messages to be in increasing order,
        according to the exit code constans above.
        """
        missing_producer = (
            "Model/Models is configured to consume signals that are not produced. "
            "Remove signal or fix/add producer."
        )
        missing_consumer = (
            "Model/Models creates signals that are not consumed. "
            "Remove or add to <mdl>_Unconsumed.csv."
        )
        incorrect_csv = (
            "Consumed signals is defined in <mdl>_Unconsumed.csv. "
            "Remove signals or fix/remove consumer."
        )
        never_active_signals = (
            "Never active signals will not appear in generated .c file. "
            "Remove corresponding part in Simulink model, signals probablty lead to terminators."
        )
        exit_codes_messages = [
            missing_producer,
            missing_consumer,
            incorrect_csv,
            never_active_signals,
        ]
        exit_code_bit_field = [
            bool(int(b)) for b in bin(exit_code)[2:]
        ]  # [2:] gets rid of 0b part
        for idx, value in enumerate(exit_code_bit_field):
            if value:
                LOGGER.info(exit_codes_messages[idx])

    def aggregate_model_inports(self):
        """Aggregate inport signals from models"""
        used_inports = set()
        for _, unit_cfg in self.per_unit_cfgs.items():
            for _, inp in unit_cfg.items():
                used_inports = used_inports.union(set(inp["inports"].keys()))
        return used_inports

    def aggregate_supplier_inports(self):
        """Aggregate outports from signal interface (supplier inports)"""
        external_outports = set()
        for if_output in self.signal_ifs.values():
            for signals in if_output.get_external_outputs().values():
                external_outports = external_outports.union(set(signals.keys()))
        return external_outports

    def run(self, args):
        """Run signal inconsistency check."""
        exit_code = EXIT_CODE_OK
        if self.models:
            prj_cfs = get_project_configs(args.project_folders)
            self.signal_ifs, self.per_unit_cfgs = get_signal_interfaces(
                prj_cfs, self.models
            )
            self.never_active_signals = get_never_active_signals(prj_cfs, self.models)
            LOGGER.info("Start check signal inconsistencies for: %s", self.models)
            self.signal_inconsistency_results = check_inconsistency(
                self.signal_ifs, self.never_active_signals
            )
            exit_code, log_output = self._analyse_inconsistency()
            LOGGER.info("Finished check signal inconsistencies %s", log_output)
            self._print_console_log_info(exit_code)
            if args.report:
                LOGGER.info(
                    "Start generating interface inconsistencies html-report for all projects"
                )
                gen_sig_incons_index_file(
                    list(self.signal_inconsistency_results.keys())
                )
                self._create_reports()
                LOGGER.info("Finished - generating inconsistencies html-reports")
        else:
            LOGGER.info("No models in change")
        return exit_code


def run_signal_inconsistency_check(args: argparse.Namespace) -> int:
    """Create Signal Inconsistency instance and run checks."""
    sig_in = SignalInconsistency(args)
    return sig_in.run(args)


def main(argv: Optional[List[str]] = None) -> int:
    """Create Signal Inconsistency instance and run checks."""
    parser = argparse.ArgumentParser(description=PARSER_HELP)
    configure_parser(parser)
    args = parser.parse_args(argv)
    return args.func(args)


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Script to update configs based on c-files."""
import argparse
import copy
import glob
import itertools
import json
import operator
import os
import re
import sys
from pprint import pformat
from typing import List, Optional

from powertrain_build.lib import logger

LOGGER = logger.create_logger('config')


class ConfigParserCommon:
    """Parser for c and h files."""

    def __init__(self):
        """Initialize common properties."""
        self.ifs = []
        self.def_map = {}
        self.configs = {}
        self.code_regexes = [(re.compile(r'^\s*#(?P<type>if|ifdef|ifndef) (?P<condition>.*)$'),
                              self.parse_if),
                             (re.compile(r'^\s*#else.*$'), self.parse_else),
                             (re.compile(r'^\s*#define (\w*)\s?(.*)?'), self.parse_defines),
                             (re.compile(r'^\s*#endif.*$'), self.parse_endif)]

    def parse_line(self, line):
        """Process each regex.

        Arguments:
        line (str): line of code
        """
        for regex, function in self.code_regexes:
            self.process_regex(line, regex, function)

    @staticmethod
    def process_regex(line, regex, function):
        """Process one regex.

        Arguments:
        line (str): line of code
        regex (object): compiled re object
        function (function): function to run if regex matches
        """
        match = regex.match(line)
        if match:
            function(*match.groups())

    def parse_file_content(self, file_content):
        """Parse each line in the file.

        Arguments:
        file_contents (list): Contents of a file
        """
        for line in file_content:
            self.parse_line(line)

    def parse_if(self, if_type, condition):
        """Parse an if-preprocessor statement.

        Arguments:
        match (object): match object
        """
        self.ifs.append((if_type, condition))

    def parse_else(self):
        """Stub for parsing."""
        raise NotImplementedError

    def parse_defines(self, variable, definition):
        """Stub for parsing."""
        raise NotImplementedError

    def parse_endif(self):
        """Parse an endif-preprocessor statement.

        Arguments:
        match (object): match object
        """
        if self.ifs:
            c_type, condition = self.ifs.pop()
            LOGGER.debug('Removing %s %s', c_type, condition)

    @staticmethod
    def read_file(c_file):
        """Read file.

        Arguments:
        c_file (str): Full path to a file
        """
        file_content = ''
        with open(c_file, encoding='latin-1') as file_handle:
            for line in file_handle:
                file_content += line
        out = re.sub(r'/\*.*?\*/', '', file_content, flags=re.S).splitlines()
        return out

    @staticmethod
    def compose_and(conditions):
        """Return and conditions."""
        return f"({' && '.join(conditions)})"

    @staticmethod
    def compose_or(conditions):
        """Return and conditions."""
        return f"({' || '.join(conditions)})"

    @staticmethod
    def sort_u(item):
        """Get a unique list of configs.

        Can handle unhashable elements.

        Arguments:
        item (list): list to unique elements of.
        """
        return map(
            operator.itemgetter(0),
            itertools.groupby(sorted(item)))


class CConfigParser(ConfigParserCommon):
    """Parser for c-files."""

    def set_regexes(self, variable):
        """Create regexes to find configs for a single variable.

        Arguments:
        variable (str): variable
        """
        self.code_regexes.append((re.compile(r'.*\b({})\b.*'.format(variable)), self.parse_code))

    def parse_defines(self, variable, definition):
        """Parse defines in c-files."""
        if definition:
            LOGGER.warning('Configuration using %s might be wrong. Set to %s in the c-file', variable, definition)
        if variable:
            if self.ifs and variable == self.ifs[-1][-1]:
                self.ifs.pop()

    def parse_else(self):
        """Parse defines in c-files."""
        if_type, condition = self.ifs.pop()
        self.ifs.append(('else' + if_type, condition))

    def parse_code(self, variable):
        """Parse a line with the variable we are looking for.

        Arguments:
        match (object): match object
        """
        LOGGER.debug('Found %s with %s', variable, self.ifs)
        if variable not in self.configs:
            self.configs[variable] = []
        # In this case, we add to a list which should be joined by 'and'
        self.configs[variable].append(copy.deepcopy(self.ifs))

    @staticmethod
    def define_config(condition, header_map, ctype):
        """Get a config from the header map.

        Arguments:
        condition
        header_map
        ctype
        """
        if ctype == 'ifdef':
            if condition in header_map.keys():
                config = header_map[condition]
            else:
                config = 'ALWAYS_ACTIVE'
                LOGGER.error('Define not found: %s %s in %s', ctype, condition, header_map)
        if ctype == 'ifndef':
            if condition in header_map.keys():
                config = '!(' + header_map[condition] + ')'
            else:
                config = 'NEVER_ACTIVE'
                LOGGER.error('Define not found: %s %s in %s', ctype, condition, header_map)
        if ctype == 'elseifdef':
            if condition in header_map.keys():
                config = f'!({header_map[condition]})'
            else:
                config = 'NEVER_ACTIVE'
                LOGGER.error('Define not found: %s %s in %s', ctype, condition, header_map)
        if ctype == 'elseifndef':
            if condition in header_map.keys():
                config = header_map[condition]
            else:
                config = 'ALWAYS_ACTIVE'
                LOGGER.error('Define not found: %s %s in %s', ctype, condition, header_map)
        return config

    def get_configs(self, variable, header_map):
        """Get configs.

        Does not remove redundant configs.
        """
        configs = []
        if variable not in self.configs:
            LOGGER.warning('%s not found. Inport that leads to terminal suspected.', variable)
            return '(NEVER_ACTIVE)'
        for config in self.configs[variable]:
            tmp_config = []
            for ctype, condition in config:
                if ctype == 'if':
                    if condition in header_map.keys():
                        LOGGER.debug('Redefining %s as %s', condition, header_map[condition])
                        tmp_config.append(header_map[condition])
                    else:
                        tmp_config.append(condition)
                elif ctype == 'elseif':
                    if condition in header_map.keys():
                        LOGGER.debug('Redefining %s as !(%s)', condition, header_map[condition])
                        tmp_config.append('!(' + header_map[condition] + ')')
                    else:
                        LOGGER.debug('Negating %s to !(%s)', condition, condition)
                        tmp_config.append('!(' + condition + ')')
                else:
                    tmp_config.append(self.define_config(condition, header_map, ctype))
            if not tmp_config and config:
                LOGGER.warning('Config not found: %s from %s', config, self.configs)
                tmp_config.append('ALWAYS_ACTIVE')
                LOGGER.info('Current config: %s', tmp_config)
            elif not config:
                LOGGER.debug('No config, always active')
                tmp_config.append('ALWAYS_ACTIVE')
            configs.append(self.compose_and(list(self.sort_u(tmp_config))))
        return self.compose_or(list(self.sort_u(configs)))


class JsonConfigHandler:
    """Handle the json config."""

    def __init__(self, cparser, header_map):
        """Initialize handling of one json file.

        Arguments:
        parser (obj): c-parser
        header_map (dict): defines in the header files
        """
        self.cparser = cparser
        self.header_map = header_map

    def traverse_unit(self, struct, setup=True):
        """Go through a data structure and look for configs to update.

        Arguments:
        struct (dict): data to go through
        parser (obj): parsing object
        header_map (dict): dict of defines
        setup (bool): Set up the parser obecjt (True) or replace configs (False)
        """
        for name, data in struct.items():
            if isinstance(data, dict) and name != 'API_blk':
                # Core data has the propety config, not configs
                if data.get('API_blk') is not None or data.get('configs') is not None:
                    if setup:
                        self.cparser.set_regexes(name)
                    else:
                        data['configs'] = self.cparser.get_configs(name, self.header_map)
                else:
                    self.traverse_unit(data, setup)

    def update_config(self, struct, c_code, header_map=None):
        """Update dict.

        Arguments:
        data (dict): A configuration dict or subdict
        c_code (list): code part of a c-file
        """
        if header_map is None:
            header_map = {}
        # Set up regexes:
        self.traverse_unit(struct, setup=True)
        self.cparser.parse_file_content(c_code)
        self.traverse_unit(struct, setup=False)

    @staticmethod
    def read_config(config_file):
        """Read config file.

        Arguments:
        config_file (str): Full path to config file
        """
        with open(config_file, encoding='latin-1') as unit_json:
            unit_config = json.load(unit_json)
        return unit_config

    @staticmethod
    def write_config(config_file, unit_config):
        """Write config file.

        Arguments:
        config_file (str): Full path to config file
        unit_config (dict): Unit config to write to file
        """
        with open(config_file, 'w', encoding="utf-8") as unit_json:
            unit_json.write(json.dumps(unit_config, indent=2))


class HeaderConfigParser(ConfigParserCommon):
    """Parser for c-files."""

    def set_defines(self, defines):
        """Set already defined defines."""
        self.def_map = defines

    def parse_else(self):
        """Crash if this is found in a header."""
        raise NotImplementedError

    def parse_defines(self, variable, definition):
        """Parse defines in c-files."""
        if self.ifs and self.ifs[-1][0] == 'ifndef' and variable == self.ifs[-1][-1]:
            # We have encountered a case of:
            #
            # #ifndef a
            # #define a
            # #define b
            #
            # Then we don't want b to be dependent on a not being defined.

            c_type, condition = self.ifs.pop()
            LOGGER.debug('Removing now defined %s from ifs: %s %s', variable, c_type, condition)
        if definition:
            LOGGER.info('Redefining %s as %s', variable, definition)
            # Here we ignore the potential #if statements preceding this.
            # Have not encountered a case where that matters.
            # This structure does not support that logic.
            # Potential for bugs.
            self.configs[variable] = [definition]
        elif self.ifs:
            config = self.get_configs(self.ifs, self.def_map)
            LOGGER.info('Defining %s as %s', variable, config)
            if variable not in self.configs:
                self.configs[variable] = []
            self.configs[variable].append(copy.deepcopy(config))
            self.def_map.update({variable: copy.deepcopy(config)})

    @staticmethod
    def define_config(condition, header_map, ctype):
        """Get a config from the header map.

        Arguments:
        condition
        header_map
        ctype
        """
        if ctype == 'ifdef':
            if condition in header_map.keys():
                LOGGER.debug('returning %s as %s', condition, header_map[condition])
                config = header_map[condition]
            else:
                config = 'ALWAYS_ACTIVE'
                LOGGER.warning('Not Implemented Yet: %s %s', ctype, condition)
        if ctype == 'ifndef':
            if condition in header_map.keys():
                LOGGER.debug('returning %s as %s', condition, header_map[condition])
                config = '!(' + header_map[condition] + ')'
            else:
                config = 'ALWAYS_ACTIVE'
                LOGGER.warning('Not Implemented Yet: %s %s', ctype, condition)
        return config

    def process_config(self, inconfigs, header_map):
        """Process configs."""
        configs = []
        for config in inconfigs:
            LOGGER.debug('Current config: %s', config)
            if isinstance(config, list):
                configs.append(self.process_config(config, header_map))
            else:
                ctype, condition = config
                if ctype == 'if':
                    if condition in header_map.keys():
                        configs.append(header_map[condition])
                    else:
                        configs.append(condition)
                elif ctype in ['ifdef', 'ifndef']:
                    configs.append(self.define_config(condition, header_map, ctype))
                else:
                    LOGGER.error('Not Implemented: %s', ctype)
            if not configs:
                configs = ['ALWAYS_ACTIVE']
        return list(self.sort_u(configs))

    def get_configs(self, configs, header_map):
        """Get configs.

        Does not remove redundant configs.
        """
        configs = self.process_config(configs, header_map)
        if len(configs) > 1:
            return self.compose_and(list(self.sort_u(configs)))
        return configs[0]

    def get_config(self):
        """Get the header map."""
        header_map = self.def_map
        for header_def, configs in self.configs.items():
            header_map[header_def] = self.compose_or(configs)
        return header_map


class ProcessHandler:
    """Class to collect functions for the process."""

    PARSER_HELP = "Parse configs.json and c-files, to update code switch configs"

    @staticmethod
    def configure_parser(parser: argparse.ArgumentParser):
        """Parse arguments."""
        parser.description = "Parse configs.json and c-files, to update code switch configs"

        subparser = parser.add_subparsers(
            title='Operation mode',
            dest='mode',
            help="Run chosen files on in a number of directories",
        )
        dir_parser = subparser.add_parser(
            'models',
            help="Run for one or multiple models. Script finds files generated from the model(s).")
        dir_parser.add_argument('models', nargs='+',
                                help="Space separated list of model directories")

        file_parser = subparser.add_parser('files',
                                           help="Choose specific files. Mainly for manually written configs.")
        file_parser.add_argument('c_file',
                                 help="Full path to C-file")
        file_parser.add_argument('config_file',
                                 help="Full path to config file")
        file_parser.add_argument('--aux_file',
                                 help="Full path to tl_aux file. (Optional) ")
        file_parser.add_argument('--local_file',
                                 help="Full path to OPort file. (Optional) ")

        parser.set_defaults(func=ProcessHandler.main)

    @staticmethod
    def get_files(model_path):
        """Get file paths from model path.

        Arguments:
        model_path (str): Path to a model (.mdl)

        Returns:
        local_file (str): Path to model_OPortMvd_LocalDefs.h
        aux_file (str): Path to tl_aux_defines_model.h
        config_file (str): Path to config_model.json
        c_file (str): Path to model.c

        """
        model_dir = os.path.dirname(model_path)
        LOGGER.info('Processing %s', model_dir)
        model_name = os.path.basename(model_dir)
        local_file = os.path.join(model_dir, 'pybuild_src', f'{model_name}_OPortMvd_LocalDefs.h')
        # aux_file does not contain the whole model-name if it is too long.
        aux_file = os.path.join(model_dir, 'pybuild_src', f'tl_aux_defines_{model_name[2:12]}.h')
        config_file = os.path.join(model_dir, 'pybuild_cfg', f'config_{model_name}.json')
        clean_model_name = model_name.split('__')[0]
        c_file = os.path.join(model_dir, 'pybuild_src', f'{clean_model_name}.c')
        return local_file, aux_file, c_file, config_file

    @staticmethod
    def update_config_file(c_file, config_file, header_map):
        """Update one config file.

        Arguments:
        c_file (str): Full path to c-file
        config_file (str): Full path to config.json
        oport_file (str): Full path to OPortMvd_LocalDefs.h (Optional)
        """
        LOGGER.info('Updating %s based on %s', config_file, c_file)
        cparser = CConfigParser()
        c_code = cparser.read_file(c_file)
        json_handler = JsonConfigHandler(cparser, header_map)
        unit_config = json_handler.read_config(config_file)
        json_handler.update_config(unit_config, c_code, header_map)
        json_handler.write_config(config_file, unit_config)

    @staticmethod
    def get_header_config(header_file, def_map):
        """Get header config.

        Arguments:
        c_file (str): Full path to c-file
        config_file (str): Full path to config.json
        oport_file (str): Full path to OPortMvd_LocalDefs.h (Optional)
        """
        if header_file is None:
            LOGGER.info('File not found: %s', header_file)
            return def_map
        if not os.path.isfile(header_file):
            LOGGER.info('File not found: %s', header_file)
            model_dir = os.path.dirname(header_file)
            for tl_aux_file in glob.glob(os.path.join(model_dir, 'tl_aux*')):
                LOGGER.warning('Looking for %s?', tl_aux_file)
            return def_map
        LOGGER.info('Parsing %s', header_file)
        parser = HeaderConfigParser()
        header_code = parser.read_file(header_file)
        parser.set_defines(def_map)
        parser.parse_file_content(header_code)
        LOGGER.debug('Header configs: %s', pformat(parser.configs))
        return parser.get_config()

    @classmethod
    def main(cls, args: argparse.Namespace):
        """Run the main function of the script."""
        if args.mode == 'files':
            LOGGER.info('Using manually supplied files %s', args)
            local_defs = cls.get_header_config(args.local_file, {})
            aux_defs = cls.get_header_config(args.aux_file, local_defs)
            cls.update_config_file(args.c_file, args.config_file, aux_defs)
        else:
            for model in args.models:
                local_file, aux_file, c_file, config_file = cls.get_files(model)
                local_defs = cls.get_header_config(local_file, {})
                aux_defs = cls.get_header_config(aux_file, local_defs)
                if os.path.isfile(c_file) and os.path.isfile(config_file):
                    cls.update_config_file(c_file, config_file, aux_defs)


def main(argv: Optional[List[str]] = None):
    """Run main function."""
    parser = argparse.ArgumentParser(ProcessHandler.PARSER_HELP)
    ProcessHandler.configure_parser(parser)
    args = parser.parse_args(argv)
    return args.func(args)


if __name__ == "__main__":
    main(sys.argv[1:])



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module for generation of c- and a2l-file with dummy signal declarations."""

import powertrain_build.build_defs as bd

from powertrain_build.types import byte_size_string, get_bitmask
from powertrain_build.a2l import A2l
from powertrain_build.problem_logger import ProblemLogger


class DummyVar(ProblemLogger):
    """Generate c- and a2l-files which declares all missing variables in the interfaces.

    TODO: Please remove this file! Only used while testing.
    """

    def __init__(self, unit_cfg, ext_dict, res_dict, prj_cfg, user_defined_types):
        """Initialize instance of class."""
        super().__init__()
        self._unit_cfg = unit_cfg
        self._unit_vars = unit_cfg.get_per_cfg_unit_cfg()
        self._ext_dict = ext_dict
        self._res_dict = res_dict
        self._ext_vars = {}
        self._int_vars = {}
        self._prj_cfg = prj_cfg
        self._enumerations = user_defined_types.get_enumerations()
        self._common_header_files = user_defined_types.common_header_files

    def _get_byte_size_string(self, data_type):
        """Get byte size of a data type as string.
        Enumeration byte sizes are derived from the underlying data type.

        Args:
            data_type (str): Data type.
        Returns:
            byte_size_string(powertrain_build.types.byte_size_string): Return result of
                                                                       powertrain_build.types.byte_size_string.
        """
        if data_type in self._enumerations:
            return byte_size_string(self._enumerations[data_type]["underlying_data_type"])
        return byte_size_string(data_type)

    def _restruct_input_data(self):
        """Restructure all the input variables per data-type.

        This will be used for declaring the variables and generating the
        A2L-file
        """
        ext_out = {
            var: data
            for ioclass, vardict in self._ext_dict.items()
            if ioclass.endswith("-Output")
            for var, data in vardict.items()
        }
        ext_ = {}
        for var in self._res_dict["sigs"]["ext"]["missing"]:
            self.debug("ext: %s", var)
            if var in ext_out:
                data = ext_out[var]
                self.debug("ext_data: %s", data)
                ext_[var] = data
        int_ = {}
        for unit in self._res_dict["sigs"]["int"]:
            for var in self._res_dict["sigs"]["int"][unit]["missing"]:
                if var not in ext_ and var in self._unit_vars["inports"]:
                    data = self._unit_vars["inports"][var][unit]
                    int_[var] = data
        for var, data in int_.items():
            data_type_size = self._get_byte_size_string(data["type"])
            self._int_vars.setdefault(data_type_size, {})[var] = data
        for var, data in ext_.items():
            data_type_size = self._get_byte_size_string(data["type"])
            self._ext_vars.setdefault(data_type_size, {})[var] = data

    def _a2l_dict(self):
        """Return a dict defining all parameters for a2l-generation."""
        res = {"vars": {}, "function": "VcDummy"}
        for inp in [self._ext_vars]:
            for sizes in inp.values():
                for var, data in sizes.items():
                    if data["type"] in self._enumerations:
                        data_type = self._enumerations[data["type"]]["underlying_data_type"]
                    else:
                        data_type = data["type"]

                    resv = res["vars"]
                    resv.setdefault(var, {})["a2l_data"] = {
                        "bitmask": get_bitmask(data_type),
                        "description": data.get("description", ""),
                        "lsb": "2^0",
                        "max": data.get("max"),
                        "min": data.get("min"),
                        "offset": "0",
                        "unit": data["unit"],
                        "x_axis": None,
                        "y_axis": None,
                    }
                    resv[var]["array"] = []
                    resv[var]["function"] = ["VcEc"]
                    resv[var]["var"] = {"cvc_type": "CVC_DISP", "type": data_type, "var": var}
        return res

    @classmethod
    def _generate_var_defs(cls, fh_c, vars, enums, comment):
        """Generate the variable definitions."""
        fh_c.write(f"\n{comment}\n\n")
        for varsize in sorted(vars.keys(), reverse=True):
            fh_c.write(f"/* Variables of size {varsize} bytes */\n\n")
            var_defs = vars[varsize]
            for var in sorted(var_defs.keys()):
                data = var_defs[var]
                if data["type"] in enums:
                    if enums[data["type"]]["default_value"] is not None:
                        init_value = enums[data["type"]]["default_value"]
                    else:
                        cls.warning('Initializing enumeration %s to "zero".', data["type"])
                        init_value = [k for k, v in enums[data["type"]]["members"].items() if v == 0][0]
                    fh_c.write(f"{data['type']} {var} = {init_value};\n")
                else:
                    fh_c.write(f"{data['type']} {var} = {0};\n")
        fh_c.write("\n")

    @classmethod
    def _generate_var_initialization(cls, fh_c, vars, comment):
        """Generate the variable initializations."""
        fh_c.write(f"\n{comment}\n\n")
        fh_c.write("\nvoid RESTART_VcDummy(void)\n{\n")
        for varsize in sorted(vars.keys(), reverse=True):
            var_defs = vars[varsize]
            for var in sorted(var_defs.keys()):
                fh_c.write(f"    {var} = {0};\n")
        fh_c.write("}\n")

    def _generate_c_file(self, filename):
        """Generate the c-file defining all missing input variables."""
        general_includes = ""
        general_includes += self._unit_cfg.base_types_headers
        for common_header_file in self._common_header_files:
            general_includes += f'#include "{common_header_file}"\n'
        general_includes += "\n"

        with open(filename, "w", encoding="utf-8") as fh_c:
            fh_c.write(general_includes)
            fh_c.write(f'#include "{bd.CVC_DISP_START}"\n\n')
            self._generate_var_defs(fh_c, self._ext_vars, self._enumerations, "/** Missing external signals **/")
            fh_c.write(f'\n#include "{bd.CVC_DISP_END}"\n')
            self.info("Generated %s", filename)

    def generate_file(self, filename):
        """Generate the files for defining all missing input variables."""
        self._restruct_input_data()
        self._generate_c_file(filename + ".c")
        a2l_dict = self._a2l_dict()
        A2l(a2l_dict, self._prj_cfg).gen_a2l(filename + ".a2l")



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Version compatibility check."""

import re

from powertrain_build import __config_version__


class Version:
    """Class encapsulating version numbers."""

    _RE = re.compile(r'^([0-9]+)\.([0-9]+)\.([0-9]+)$')
    _app_version = (0, 0, 0)

    def __init__(self, version):
        """Init."""
        if isinstance(version, float):
            version = f'{version}.0'
        elif isinstance(version, int):
            version = f'{version}.0.0'
        if isinstance(version, str):
            match = Version._RE.match(version)
            if match:
                self._version = (int(match.group(1)), int(match.group(2)), int(match.group(3)))
            else:
                raise ValueError(f'Not a valid version string: {version}')
        else:
            raise ValueError(f'Not a valid version: {version}')

    def __len__(self):
        """Get length of object."""
        return 3

    def __getitem__(self, i):
        """Get item with index i."""
        return self._version[i]

    @classmethod
    def is_compatible(cls, version):
        """Check version compatibility.

        Args:
            version (str/int/float): The version to be compared with the version of powertrain_build.

        Returns:
            bool: True/False based on version compatibility.

        Raises:
            ValueError: If the provided version is not a valid version string, integer, or float.
        """
        version_is_compatible = False
        if version is not None:
            version_cls = Version(version)

            check_major = cls._app_version[0] == version_cls[0]
            check_minor = cls._app_version[1] >= version_cls[1]
            check_patch = cls._app_version[2] >= version_cls[2]

            version_is_compatible = check_major and check_minor and check_patch

        return version_is_compatible


Version._app_version = Version(__config_version__)



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-
"""Module containing classes for generating RTE dummy code.

These files are needed for building test SW,
where the RTE is not available.
For example, when running Silver tests.
"""

from powertrain_build.problem_logger import ProblemLogger

from pathlib import Path


class RteDummy(ProblemLogger):
    """A class for RTE dummy file generation."""

    def __init__(self, build_cfg, core, nvm_def, calib_data):
        """Init.

        Args:
            build_cfg (BuildProjConfig): Object with build configuration settings.
            core (ZCCore): Object representing the core configuration data.
            nvm_def (ZCNVMDef): Object with NVM definition information.
            calib_data (dict): Dictionary containing calibration data for a ZoneController project.
        """
        super().__init__()
        self.build_cfg = build_cfg
        self.dtcs = core.project_dtcs
        self.nvm_def = nvm_def
        self.calibration_variables = calib_data.get("class_info", {})
        mem_map_config = self.build_cfg.get_memory_map_config()
        self.mem_map_header_file_name = f"{mem_map_config['memMapPrefix']}_MemMap"
        self.swc_header_file_name = f"Rte_{self.build_cfg.get_composition_config('softwareComponentName')}"
        self.type_header_file_name = "Rte_Type"
        self.source_file_name = "Rte_Dummy"

    def _get_common_header(self):
        """Get common header for the RTE dummy files."""
        return (
            "/*\n"
            " * This file is generated by the Powertrain Build System.\n"
            " * It defines RTE dummy types and/or functions.\n"
            " * Do not modify this file manually.\n"
            " */\n"
        )

    def _get_mem_map_header_content(self):
        """Get content for the dummy memory map header."""
        return (
            f"{self._get_common_header()}"
            f"#ifndef {self.mem_map_header_file_name.upper()}_H\n"
            f"#define {self.mem_map_header_file_name.upper()}_H\n\n"
            f"#endif /* {self.mem_map_header_file_name.upper()}_H */\n"
        )

    def _get_swc_header_content(self):
        """Get content for the SWC RTE dummy header."""
        return (
            f"{self._get_common_header()}"
            f"#ifndef {self.swc_header_file_name.upper()}_H\n"
            f"#define {self.swc_header_file_name.upper()}_H\n\n"
            f'#include "{self.type_header_file_name}.h"\n'
            f"#endif /* {self.swc_header_file_name.upper()}_H */\n"
        )

    def _get_type_header_header(self):
        """Get header for the RTE type dummy header."""
        include_nvm_header = ""
        if not self.build_cfg.get_code_generation_config("useRteNvmStructs"):
            nvm_defs = self.build_cfg.get_nvm_defs()
            nvm_file_name = nvm_defs["fileName"]
            include_nvm_header = f'#include "{nvm_file_name}.h"\n'
        return (
            f"{self._get_common_header()}"
            f"#ifndef {self.type_header_file_name.upper()}_H\n"
            f"#define {self.type_header_file_name.upper()}_H\n\n"
            '#include "tl_basetypes.h"\n'
            f"{include_nvm_header}\n"
            "#define FALSE 0U\n"
            "#define TRUE 1U\n\n"
        )

    def _get_type_header_footer(self):
        """Get footer for the RTE type dummy header."""
        return f"\n#endif /* {self.type_header_file_name.upper()}_H */\n"

    def _get_source_header(self):
        """Get header for the RTE dummy source."""
        return (
            f"{self._get_common_header()}"
            f'#include "{self.swc_header_file_name}.h"\n\n'
        )

    def _get_typedef_dummy(self):
        """Get typedef dummy code."""
        typedefs = []

        function_name = self.build_cfg.get_composition_config("compositionName")
        if function_name is None:
            function_name = self.build_cfg.get_composition_config("softwareComponentName")

        # MAPs get typedef:ed to structs and need special data type mapping in calibration.py
        for signal_name, signal_data in self.calibration_variables.items():
            if isinstance(signal_data["width"], list):
                if signal_name.startswith("m") and signal_name[-2:] not in ["_r", "_c"]:
                    typedefs.append(
                        f"typedef {signal_data['type']} {signal_data['autosar_type']}_e[{signal_data['width'][0]}];\n"
                    )
                    typedefs.append(
                        "typedef struct \n"
                        "{\n"
                        f"  {signal_data['autosar_type']}_e {signal_data['autosar_type']}[{signal_data['width'][1]}];\n"
                        f"}} {signal_data['autosar_type']};\n\n"
                    )

        return "".join(typedefs)

    def _get_nvm_header_dummy(self):
        """Get NVM dummy header code."""
        struct_defines = []
        function_declarations = []
        prefix = self.build_cfg.get_scheduler_prefix()
        use_rte_nvm_structs = self.build_cfg.get_code_generation_config("useRteNvmStructs")

        for memory_area in self.nvm_def._nvm_memory_areas:
            nvm_name = f"{prefix}{memory_area}"
            struct_name = f"dt_{nvm_name}" if use_rte_nvm_structs else f"struct {nvm_name}"
            function_declarations.append(f"{struct_name} *Rte_Pim_{nvm_name}(void);")
            function_declarations.append(f"void Rte_Read_{nvm_name.upper()}_{nvm_name.upper()}({struct_name} *block);")
            function_declarations.append(f"void Rte_Write_{nvm_name.upper()}_{nvm_name.upper()}({struct_name} *block);")

            if use_rte_nvm_structs:
                struct_defines.append("typedef struct\n{")
                memory_area_index = self.nvm_def._get_nvm_areas_index(memory_area)
                nr_of_unused_signals = self.nvm_def.nvm_definitions[memory_area_index]["size"]
                signals = self.nvm_def.nvm_definitions[memory_area_index]["signals"]
                for signal in signals:
                    signal_string = ""
                    nr_of_unused_signals -= signal["x_size"] * signal["y_size"]
                    signal_string += f'  {signal["type"]} {self.nvm_def.struct_member_prefix}{signal["name"]}'
                    size = max(signal["x_size"], 1) * max(signal["y_size"], 1)
                    if size > 1:
                        if signal["x_size"] > 1:
                            signal_string += f'[{signal["x_size"]}]'
                        if signal["y_size"] > 1:
                            signal_string += f'[{signal["y_size"]}]'
                    signal_string += ";"
                    struct_defines.append(signal_string)
                if nr_of_unused_signals > 0:
                    struct_defines.append(
                        f'  {self.nvm_def.nvm_definitions[memory_area_index]["default_datatype"]} '
                        f'unused[{nr_of_unused_signals}];'
                    )
                struct_defines.append(f"}} {struct_name};\n")

        return "\n".join(struct_defines + function_declarations)

    def _get_nvm_source_dummy(self):
        """Generate NVM source dummy code."""
        lines_to_write = []
        prefix = self.build_cfg.get_scheduler_prefix()
        use_rte_nvm_structs = self.build_cfg.get_code_generation_config("useRteNvmStructs")
        for memory_area in self.nvm_def._nvm_memory_areas:
            nvm_name = f"{prefix}{memory_area}"
            struct_name = f"dt_{nvm_name}" if use_rte_nvm_structs else f"struct {nvm_name}"
            return_string = f"{struct_name} dummy; {struct_name}* dummyPtr = &dummy; return dummyPtr;"
            lines_to_write.append(f"{struct_name} *Rte_Pim_{nvm_name}(void) {{ {return_string} }}")
            lines_to_write.append(f"void Rte_Read_{nvm_name.upper()}_{nvm_name.upper()}({struct_name} *block) {{}}")
            lines_to_write.append(f"void Rte_Write_{nvm_name.upper()}_{nvm_name.upper()}({struct_name} *block) {{}}")
        lines_to_write.append("")
        return "\n".join(lines_to_write)

    def _get_calibration_function_data(self, signal_name, signal_data):
        """"Get calibration function data.

        Args:
            signal_name (str): Name of the signal.
            signal_data (dict): Dictionary containing signal data.
        Returns:
            return_type, return_string (tuple): Tuple containing the return type and return string.
        """
        if isinstance(signal_data["width"], list):
            if signal_data["width"][0] == 1:
                dummy_variable = f"dummy[{signal_data['width'][1]}]"
            else:
                dummy_variable = f"dummy[{signal_data['width'][0]}][{signal_data['width'][1]}]"
            # MAPs get typedef:ed to structs and need special data type mapping in calibration.py
            if signal_name.startswith("m") and signal_name[-2:] not in ["_r", "_c"]:
                return_type = f'const {signal_data["autosar_type"]}*'
                return_string = (
                    f"const {signal_data['autosar_type']} {dummy_variable}; "
                    f"{return_type} dummyPtr = &dummy; "
                    "return dummyPtr;"
                )
            else:
                return_type = f'const {signal_data["type"]}*'
                return_string = (
                    f"{signal_data['type']} {dummy_variable}; "
                    f"{return_type} dummyPtr = &dummy; "
                    "return dummyPtr;"
                )
        else:
            return_type = f'{signal_data["type"]}'
            return_string = f"return ({return_type})0;"
        return return_type, return_string

    def _get_calibration_header_dummy(self):
        """Get calibration dummy header code."""
        if not self.calibration_variables:
            return ""

        lines_to_write = []
        swc_name = self.build_cfg.get_composition_config("softwareComponentName")
        for signal_name, signal_data in self.calibration_variables.items():
            return_type = self._get_calibration_function_data(signal_name, signal_data)[0]
            lines_to_write.append(f"{return_type} Rte_CData_{swc_name}_{signal_name}(void);")
        lines_to_write.append("")
        return "\n".join(lines_to_write)

    def _get_calibration_source_dummy(self):
        """Get calibration source dummy code."""
        if not self.calibration_variables:
            return ""

        lines_to_write = []
        swc_name = self.build_cfg.get_composition_config("softwareComponentName")
        for signal_name, signal_data in self.calibration_variables.items():
            return_type, return_string = self._get_calibration_function_data(signal_name, signal_data)
            lines_to_write.append(f"{return_type} Rte_CData_{swc_name}_{signal_name}(void) {{ {return_string} }}")
        lines_to_write.append("")
        return "\n".join(lines_to_write)

    def _get_core_header_dummy(self):
        """Get core header dummy code."""
        if not self.dtcs:
            return ""

        lines_to_write = []
        for dtc in self.dtcs:
            lines_to_write.append(f"UInt8 Rte_Call_Event_{dtc}_SetEventStatus(UInt8 EventStatus);")
        lines_to_write.append("")
        return "\n".join(lines_to_write)

    def _get_core_source_dummy(self):
        """Get core source dummy code."""
        if not self.dtcs:
            return ""

        lines_to_write = []
        for dtc in self.dtcs:
            lines_to_write.append(f"UInt8 Rte_Call_Event_{dtc}_SetEventStatus(UInt8 EventStatus) {{ return 0U; }}")
        lines_to_write.append("")
        return "\n".join(lines_to_write)

    def generate_rte_dummy(self):
        """Generate RTE dummy files."""
        src_code_dest_dir = self.build_cfg.get_src_code_dst_dir()
        mem_map_header_file = Path(src_code_dest_dir, self.mem_map_header_file_name + ".h")
        swc_header_file = Path(src_code_dest_dir, self.swc_header_file_name + ".h")
        type_header_file = Path(src_code_dest_dir, self.type_header_file_name + ".h")
        source_file = Path(src_code_dest_dir, self.source_file_name + ".c")
        with mem_map_header_file.open(mode="w", encoding="utf-8") as mem_map_header_fh:
            mem_map_header_fh.write(self._get_mem_map_header_content())
        with swc_header_file.open(mode="w", encoding="utf-8") as swc_header_fh:
            swc_header_fh.write(self._get_swc_header_content())
        with type_header_file.open(mode="w", encoding="utf-8") as type_header_fh:
            type_header_fh.write(self._get_type_header_header())
            type_header_fh.write(self._get_typedef_dummy())
            type_header_fh.write(self._get_core_header_dummy())
            type_header_fh.write(self._get_nvm_header_dummy())
            type_header_fh.write(self._get_calibration_header_dummy())
            type_header_fh.write(self._get_type_header_footer())
        with source_file.open(mode="w", encoding="utf-8") as source_fh:
            source_fh.write(self._get_source_header())
            source_fh.write(self._get_core_source_dummy())
            source_fh.write(self._get_nvm_source_dummy())
            source_fh.write(self._get_calibration_source_dummy())



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-
"""Module for a2l-file generation."""

import sys
import re
import logging
from string import Template
from pprint import pformat
from powertrain_build.problem_logger import ProblemLogger
from powertrain_build.types import a2l_type, a2l_range

LOG = logging.getLogger()


class A2l(ProblemLogger):
    """Class for a2l-file generation."""

    def __init__(self, var_data_dict, prj_cfg):
        """Generate a2l-file from provided data dictionary.

        Args:
            var_data_dict (dict): dict defining all variables and parameters in a2l

        Sample indata structure:
        ::

            {
                "function": "rVcAesSupM",
                "vars": {
                    "rVcAesSupM_p_SupMonrSCTarDly": {
                        "var": {
                            "type": "Float32",
                            "cvc_type": "CVC_DISP"
                        },
                        "a2l_data": {
                            "unit": "kPa",
                            "description": "Low pass filtered supercharger target pressure",
                            "max": "300",
                            "min": "0",
                            "lsb": "1",
                            "offset": "0",
                            "bitmask": None,
                            "x_axis": None,
                            "y_axis": None,
                        },
                        "array": None
                    }
                }
            }
        """
        super().__init__()
        self._var_dd = var_data_dict
        self._prj_cfg = prj_cfg
        self._axis_ref = None
        self._axis_data = None
        self._compu_meths = None
        self._rec_layouts = None
        self._fnc_outputs = None
        self._fnc_inputs = None
        self._fnc_locals = None
        self._fnc_char = None
        # generate the a2l string
        self._gen_a2l()

    def gen_a2l(self, filename):
        """Write a2l-data to file.

        Args:
            filename (str): Name of the generated a2l-file.
        """
        with open(filename, 'w', encoding="utf-8") as a2l:
            a2l.write(self._a2lstr)
            self.debug('Generated %s', filename)

    def _gen_a2l(self):
        """Generate an a2l-file based on the supplied data dictionary."""
        self._gen_compu_methods()
        # self.debug('_compu_meths')
        # self.debug(pp.pformat(self._compu_meths))
        self._find_axis_ref()
        # self.debug("_axis_ref")
        # self.debug(pp.pformat(self._axis_ref))
        self._find_axis_data()
        self._gen_record_layouts_data()
        # self.debug("_axis_data")
        # self.debug(pp.pformat(self._axis_data))
        # self.debug("_rec_layouts")
        # self.debug(pp.pformat(self._rec_layouts))
        self._check_axis_ref()

        output_meas = ''
        output_char = ''
        output_axis = ''

        self._fnc_outputs = []
        self._fnc_inputs = []
        self._fnc_locals = []
        self._fnc_char = []

        for var, data in self._var_dd['vars'].items():
            try:
                cvc_type = data['var']['cvc_type']
                if 'CVC_DISP' in cvc_type:
                    output_meas += self._gen_a2l_measurement_blk(var, data)
                    self._fnc_locals.append(var)
                elif 'CVC_CAL' in cvc_type:
                    self._fnc_char.append(var)
                    srch = re.search('_[rcxyXY]$', var)
                    if srch is None:
                        output_char += self._gen_a2l_characteristic_blk(var, data)
                    else:
                        output_axis += self._gen_a2l_axis_pts_blk(var, data)
                elif cvc_type == 'CVC_NVM':
                    output_meas += self._gen_a2l_measurement_blk(var, data)
                    self._fnc_locals.append(var)
                elif cvc_type == 'CVC_IN':
                    self._outputs = None
                    self._inputs = None
                    self._fnc_inputs.append(var)
                elif cvc_type == 'CVC_OUT':
                    self._fnc_outputs.append(var)
            except TypeError:
                self.warning("Warning: %s has no A2l-data", var)
            except Exception as e:
                self.critical("Unexpected error: %s", sys.exc_info()[0])
                raise e
        # generate COMPU_METHS
        output_compu_m = ''
        for k in self._compu_meths:
            output_compu_m += self._gen_a2l_compu_metod_blk(k)

        # generate FUNCTIONS
        output_funcs = self._gen_a2l_function_blk()

        # generate RECORD_LAYOUTS
        output_recl = ''
        for k in self._rec_layouts:
            output_recl += self._gen_a2l_rec_layout_blk(k)

        output = output_char + output_axis + output_meas + \
            output_compu_m + output_funcs + output_recl
        # self.debug(pp.pformat(self._var_dd))
        # self.debug('Output:')
        # self.debug(output)
        self._a2lstr = output

    def _find_axis_data(self):
        """Parse all variables and identify axis points.

        TODO: Change this function to check for names with _r | _c
              suffixes
        """
        self._axis_data = {}
        variable_names = self._var_dd['vars'].keys()
        for name in variable_names:
            x_nm = y_nm = None
            if name + '_x' in variable_names:
                x_nm = name + '_x'
            if name + '_y' in variable_names:
                y_nm = name + '_x'
            if x_nm is not None or y_nm is not None:
                self._axis_data[name] = (x_nm, y_nm)

    def _find_axis_ref(self):
        """Parse all variables and identify which are defined as axis points."""
        self._axis_ref = {}

        for var, data in self._var_dd['vars'].items():
            if data.get('a2l_data') is not None:
                x_axis = data['a2l_data'].get('x_axis')
                y_axis = data['a2l_data'].get('y_axis')
                if x_axis is not None:
                    if x_axis in self._axis_ref:
                        self._axis_ref[x_axis]['used_in'].append((var, 'x'))
                    else:
                        self._axis_ref[x_axis] = {'used_in': [(var, 'x')]}
                if y_axis is not None:
                    if y_axis in self._axis_ref:
                        self._axis_ref[y_axis]['used_in'].append((var, 'y'))
                    else:
                        self._axis_ref[y_axis] = {'used_in': [(var, 'y')]}

    @classmethod
    def _get_a2d_minmax(cls, a2d, ctype=None):
        """Get min max limits from a2l data.

        Gives max limits if min/max limits are undefined.
        """
        typelim = a2l_range(ctype)
        minlim = a2d.get('min')
        if minlim is None or minlim == '-':
            minlim = typelim[0]
        maxlim = a2d.get('max')
        if maxlim is None or maxlim == '-':
            maxlim = typelim[1]
        return minlim, maxlim

    def _check_axis_ref(self):
        """Check that the axis definitions are defined in the code."""
        undef_axis = [ax for ax in self._axis_ref
                      if ax not in self._var_dd['vars']]
        if undef_axis:
            self.warning(f'Undefined axis {pformat(undef_axis)}')

    def _gen_compu_methods(self):
        """Generate COMPU_METHOD data, and add it into the var_data_dict."""
        self._compu_meths = {}
        for var, data in self._var_dd['vars'].items():
            a2d = data.get('a2l_data')
            if a2d is not None:

                lsb = self._calc_lsb(a2d['lsb'])
                offset_str = str(a2d['offset'])
                is_offset_num = bool(re.match('[0-9]', offset_str))
                if is_offset_num:
                    offset = float(offset_str)
                else:
                    offset = 0
                key = (lsb, offset, a2d['unit'])
                self._var_dd['vars'][var]['compu_meth'] = key
                name = self._compu_key_2_name(key)
                if key in self._compu_meths:
                    self._compu_meths[key]['vars'].append(var)
                else:
                    self._compu_meths[key] = {'name': name,
                                              'vars': [var],
                                              'coeffs': self._get_coefs_str(lsb,
                                                                            offset)}

    def _compu_key_2_name(self, key):
        """Generate a COMPU_METHOD name from the keys in the name.

        Args:
            key (tuple): a list with compumethod keys (lsb, offset, unit)
        """
        conversion_list = [(r'[\./]', '_'), ('%', 'percent'),
                           ('-', 'None'), (r'\W', '_')]
        name = f"{self._var_dd['function']}_{key[0]}_{key[1]}_{key[2]}"
        for frm, to_ in conversion_list:
            name = re.sub(frm, to_, name)
        return name

    @staticmethod
    def _array_to_a2l_string(array):
        """Convert c-style array definitions to A2L MATRIX_DIM style."""
        if not isinstance(array, list):
            array = [array]
        dims = [1, 1, 1]
        for i, res in enumerate(array):
            dims[i] = res
        return f"MATRIX_DIM {dims[0]} {dims[1]} {dims[2]}"

    @staticmethod
    def _get_coefs_str(lsb, offset):
        """Calculate the a2l-coeffs from the lsb and offs fields.

        The fields are defined in the a2l_data dictionary.
        """
        return f"COEFFS 0 1 {offset} 0 0 {lsb}"

    @staticmethod
    def _calc_lsb(lsb):
        """Convert 2^-2, style lsbs to numericals."""
        if isinstance(lsb, str):
            if lsb == '-':
                return 1
            shift = re.match(r'(\d+)\^([\-+0-9]+)', lsb)
            if shift is not None:
                lsb_num = pow(int(shift.group(1)), int(shift.group(2)))
            else:
                lsb_num = float(lsb)
            return lsb_num
        return lsb

    def _gen_record_layouts_data(self):
        """Generate record layouts."""
        self._rec_layouts = {}
        for var, data in self._var_dd['vars'].items():
            if data.get('a2l_data') is not None:
                a2l_unit = a2l_type(data['var']['type'])
                # if calibration data has a suffix of _x of _y it is a axis_pts
                srch = re.search('_[xyXY]$', var)
                if srch is not None:
                    name = a2l_unit + "_X_INCR_DIRECT"
                    self._rec_layouts[name] = f"AXIS_PTS_X 1 {a2l_unit} INDEX_INCR DIRECT"
                    data['rec_layout'] = name
                else:
                    name = a2l_type(data['var']['type']) + "_COL_DIRECT"
                    self._rec_layouts[name] = f"FNC_VALUES 1 {a2l_unit} COLUMN_DIR DIRECT"
                    data['rec_layout'] = name

    def _get_inpq_data(self, inp_quant):
        """Get the necessary InputQuantity parameters."""
        if inp_quant is not None:
            if inp_quant in self._var_dd['vars']:
                return inp_quant
        return 'NO_INPUT_QUANTITY'

    # Bosh template
    _meas_tmplt = Template("""
    /begin MEASUREMENT
        $Name /* Name */
        "$LongIdent" /* LongIdentifier */
        $Datatype   /* Datatype */
        $Conversion   /* Conversion */
        1   /* Resolution */
        0   /* Accuracy */
        $LowerLimit   /* LowerLimit */
        $UpperLimit   /* UpperLimit */
        $OptionalData
        ECU_ADDRESS 0x00000000
    /end MEASUREMENT
""")

    # Denso template
    _meas_tmplt_nvm = Template("""
    /begin MEASUREMENT
        $Name /* Name */
        "$LongIdent" /* LongIdentifier */
        $Datatype   /* Datatype */
        $Conversion   /* Conversion */
        1   /* Resolution */
        0   /* Accuracy */
        $LowerLimit   /* LowerLimit */
        $UpperLimit   /* UpperLimit */
        $OptionalData
    /end MEASUREMENT
""")

    def _gen_a2l_measurement_blk(self, var_name, data):
        """Generate an a2l MEASUREMENT block."""
        opt_data = 'READ_WRITE'
        a2d = data.get('a2l_data')
        if a2d is not None:
            c_type = data['var']['type']
            # if c_type == 'Bool':
            #     opt_data += '\n' + ' ' * 8 + "BIT_MASK 0x1"
            if a2d.get('bitmask') is not None:
                opt_data += '\n' + ' ' * 8 + "BIT_MASK %s" % a2d['bitmask']
            if data.get('array'):
                opt_data += '\n' + ' ' * 8 + \
                    self._array_to_a2l_string(data['array'])

            use_symbol_links = self._prj_cfg.get_code_generation_config(item='useA2lSymbolLinks')
            if a2d.get('symbol'):
                if use_symbol_links:
                    opt_data += '\n' + ' ' * 8 + 'SYMBOL_LINK "%s" %s' % (a2d['symbol'], a2d.get('symbol_offset'))
                    LOG.debug('This a2l is using SYMBOL_LINK for %s', opt_data)
                else:
                    var_name = a2d['symbol'] + '._' + var_name
                    LOG.debug('This a2l is not using SYMBOL_LINK for %s', var_name)

            dtype = a2l_type(c_type)
            minlim, maxlim = self._get_a2d_minmax(a2d, c_type)
            conv = self._compu_meths[data['compu_meth']]['name']

            if a2d.get('symbol') and use_symbol_links:
                res = self._meas_tmplt_nvm.substitute(Name=var_name,
                                                      LongIdent=a2d['description'].replace('"', '\\"'),
                                                      Datatype=dtype,
                                                      Conversion=conv,
                                                      LowerLimit=minlim,
                                                      UpperLimit=maxlim,
                                                      OptionalData=opt_data)
            else:
                res = self._meas_tmplt.substitute(Name=var_name,
                                                  LongIdent=a2d['description'].replace('"', '\\"'),
                                                  Datatype=dtype,
                                                  Conversion=conv,
                                                  LowerLimit=minlim,
                                                  UpperLimit=maxlim,
                                                  OptionalData=opt_data)
            return res
        return None

    _char_tmplt = Template("""
    /begin CHARACTERISTIC
        $Name /* Name */
        "$LongIdent" /* LongIdentifier */
        $Type   /* Datatype */
        0x00000000    /* address: $Name */
        $Deposit   /* Deposit */
        0   /* MaxDiff */
        $Conversion   /* Conversion */
        $LowerLimit   /* LowerLimit */
        $UpperLimit   /* UpperLimit */$OptionalData
    /end CHARACTERISTIC
""")

    def _gen_a2l_characteristic_blk(self, var, data):
        """Generate an a2l CHARACTERISTIC block."""
        opt_data = ''
        a2d = data.get('a2l_data')
        type_ = 'WRONG_TYPE'
        if a2d is not None:
            arr = data.get('array')
            if arr is not None:
                arr_dim = len(arr)
            else:
                arr_dim = 0
            # Check is axis_pts are defined for the axis, if not make the
            # type a VAL_BLK with matrix dimension, otherwise set
            # a CURVE or MAP type

            # If arr_dim is 0 the CHARACTERISTIC is a value
            if arr_dim == 0:
                type_ = 'VALUE'
            elif arr_dim == 1:
                x_axis_name = var + '_x'
                # Check if axis variable is defined
                if x_axis_name in self._var_dd['vars'].keys():
                    type_ = 'CURVE'
                    opt_data += self._gen_a2l_axis_desc_blk(self._get_inpq_data(a2d.get('x_axis')),
                                                            x_axis_name)
                else:
                    type_ = 'VAL_BLK'
                    opt_data += self._array_to_a2l_string(data['array'])
            elif arr_dim == 2:
                x_axis_name = var + '_x'
                y_axis_name = var + '_y'
                # Check if axis variable is defined
                nbr_def_axis = 0
                if x_axis_name in self._var_dd['vars'].keys():
                    nbr_def_axis += 1
                if y_axis_name in self._var_dd['vars'].keys():
                    nbr_def_axis += 1
                if nbr_def_axis == 2:
                    type_ = 'MAP'
                    inpq_x = self._get_inpq_data(a2d['x_axis'])
                    opt_data += self._gen_a2l_axis_desc_blk(inpq_x, x_axis_name)
                    inpq_y = self._get_inpq_data(a2d['y_axis'])
                    opt_data += self._gen_a2l_axis_desc_blk(inpq_y, y_axis_name)
                elif nbr_def_axis == 0:
                    type_ = 'VAL_BLK'
                    opt_data += self._array_to_a2l_string(data['array'])
                else:
                    self.warning(
                        'MAP %s has only one AXIS_PTS defined, shall be none or two', var)

            minlim, maxlim = self._get_a2d_minmax(a2d)

            res = self._char_tmplt.substitute(Name=var,
                                              LongIdent=a2d['description'].replace('"', '\\"'),
                                              Type=type_,
                                              Deposit=data['rec_layout'],
                                              Conversion=self._compu_meths[data['compu_meth']]['name'],
                                              LowerLimit=minlim,
                                              UpperLimit=maxlim,
                                              OptionalData=opt_data)
            return res
        self.warning("%s has no A2L-data", var)
        return None
        # Types ASCII CURVE MAP VAL_BLK VALUE

    _axis_desc_tmplt = Template("""
        /begin AXIS_DESCR
            COM_AXIS    /* Attribute */
            $inp_quant   /* InputQuantity */
            $conv  /* Conversion */
            $maxaxispts  /* MaxAxisPoints */
            $minlim  /* LowerLimit */
            $maxlim   /* UpperLimit */
            AXIS_PTS_REF $axis_pts_ref
            DEPOSIT ABSOLUTE
        /end AXIS_DESCR""")

    def _gen_a2l_axis_desc_blk(self, inp_quant, axis_pts_ref):
        """Generate an a2l AXIS_DESCR block.

        TODO: Check that the AXIS_PTS_REF blocks are defined
        """
        out = ''
        inp_quant_txt = self._get_inpq_data(inp_quant)
        axis_pts = self._var_dd['vars'][axis_pts_ref]
        conv = self._compu_meths[axis_pts['compu_meth']]['name']
        max_axis_pts = axis_pts['array'][0]
        min_lim, max_lim = self._get_a2d_minmax(axis_pts['a2l_data'])
        out += self._axis_desc_tmplt.substitute(inp_quant=inp_quant_txt,
                                                conv=conv,
                                                maxaxispts=max_axis_pts,
                                                minlim=min_lim,
                                                maxlim=max_lim,
                                                axis_pts_ref=axis_pts_ref)
        return out

    _compu_meth_tmplt = Template("""
    /begin COMPU_METHOD
        $name   /* Name */
        "$longident"    /* LongIdentifier */
        RAT_FUNC    /* ConversionType */
        "$format"   /* Format */
        "$unit" /* Unit */
        $coeffs
    /end COMPU_METHOD
""")

    def _gen_a2l_compu_metod_blk(self, key):
        """Generate an a2l COMPU_METHOD block."""
        cmeth = self._compu_meths[key]
        name = self._compu_key_2_name(key)
        out = self._compu_meth_tmplt.substitute(name=name,
                                                longident='',
                                                format='%11.3',
                                                unit=key[2],
                                                coeffs=cmeth['coeffs'])
        return out

    _axis_tmplt = Template("""
    /begin AXIS_PTS
        $name   /* Name */
        "$longident"   /* LongIdentifier */
        0x00000000
        NO_INPUT_QUANTITY   /* InputQuantity */
        $deposit  /* Deposit */
        0   /* MaxDiff */
        $convert  /* Conversion */
        $max_ax_pts  /* MaxAxisPoints */
        $minlim   /* LowerLimit */
        $maxlim   /* UpperLimit */
        DEPOSIT ABSOLUTE
    /end AXIS_PTS
""")

    def _gen_a2l_axis_pts_blk(self, var, data):
        """Generate an a2l AXIS_PTS block."""
        deposit = data['rec_layout']
        conv = self._compu_meths[data['compu_meth']]['name']
        max_axis_pts = data['array'][0]
        min_lim, max_lim = self._get_a2d_minmax(data['a2l_data'])
        out = self._axis_tmplt.substitute(name=var,
                                          longident=data['a2l_data']['description'],
                                          deposit=deposit,
                                          convert=conv,
                                          max_ax_pts=max_axis_pts,
                                          minlim=min_lim,
                                          maxlim=max_lim)
        return out

    _rec_layout_tmplt = Template("""
    /begin RECORD_LAYOUT
        $name /* Name */
        $string
    /end RECORD_LAYOUT
""")

    def _gen_a2l_rec_layout_blk(self, key):
        """Generate an a2l AXIS_PTS block."""
        string = self._rec_layouts[key]
        out = self._rec_layout_tmplt.substitute(name=key,
                                                string=string)
        return out

    def _gen_a2l_function_blk(self):
        """Generate an a2l FUNCTION block."""
        out = '\n    /begin FUNCTION\n'
        out += f'        {self._var_dd["function"]} /* Name */\n'
        out += '        ""  /* LongIdentifier */\n'
        if self._fnc_char:
            out += '        /begin DEF_CHARACTERISTIC\n'
            for idf in self._fnc_char:
                out += f'            {idf} /* Identifier */\n'
            out += '        /end DEF_CHARACTERISTIC\n'
        if self._fnc_inputs:
            out += '        /begin IN_MEASUREMENT\n'
            for idf in self._fnc_inputs:
                out += f'            {idf} /* Identifier */\n'
            out += '        /end IN_MEASUREMENT\n'
        if self._fnc_locals:
            out += '        /begin LOC_MEASUREMENT\n'
            for idf in self._fnc_locals:
                out += f'            {idf} /* Identifier */\n'
            out += '        /end LOC_MEASUREMENT\n'
        if self._fnc_outputs:
            out += '        /begin OUT_MEASUREMENT\n'
            for idf in self._fnc_outputs:
                out += f'            {idf} /* Identifier */\n'
            out += '        /end OUT_MEASUREMENT\n'
        out += '    /end FUNCTION\n'
        return out



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-
"""Python module used for calculating interfaces for CSP HI"""
import argparse
import sys
from pathlib import Path
from typing import List, Optional

from powertrain_build.interface import generation_utils
from powertrain_build.interface.device_proxy import DPAL
from powertrain_build.lib.helper_functions import recursive_default_dict, to_normal_dict

OP_READ = 'read'
OP_WRITE = 'write'

PARSER_HELP = "Generate HI YAML interface file."


def generate_hi_interface(args, hi_interface):
    """Generate HI YAML interface file.

    Args:
        args (Namespace): Arguments from command line.
        hi_interface (dict): HI interface dict based on HIApplication and generation_utils.get_interface.
    Returns:
        result (dict): Aggregated signal information as a dict.
    """

    io_translation = {
        'consumer': OP_READ,
        'producer': OP_WRITE
    }
    result = recursive_default_dict()
    for raster_data in hi_interface.values():
        for direction, signals in raster_data.items():
            hi_direction = io_translation[direction]
            for signal in signals:
                domain = signal['domain']
                group = signal['group']
                name = signal['variable']
                property_name = signal['property']
                if group is None:
                    port_name = signal['port_name'] or property_name
                    result[hi_direction]['signals'][domain][port_name]['data'][property_name] = name
                else:
                    port_name = signal['port_name'] or group
                    result[hi_direction]['signal_groups'][domain][port_name][group]['data'][property_name] = name
    generation_utils.write_to_file(to_normal_dict(result), args.output, is_yaml=True)


def generate_hi_interface_cli(args: argparse.Namespace):
    """CLI entrypoint for generating HI YAML interface file.

    Args:
        args (Namespace): Arguments from command line.
    """
    app = generation_utils.process_app(args.config)
    hi_app = DPAL(app)
    interface = generation_utils.get_interface(app, hi_app)
    generate_hi_interface(args, interface)


def configure_parser(parser: argparse.ArgumentParser):
    """Configure parser for generating HI YAML interface file."""
    generation_utils.add_base_args(parser)
    parser.add_argument(
        "output",
        help="Output file with interface specifications.",
        type=Path
    )
    parser.set_defaults(func=generate_hi_interface_cli)


def main(argv: Optional[List[str]] = None):
    """ Main function for stand alone execution.
    Mostly useful for testing and generation of dummy hal specifications.
    """
    parser = argparse.ArgumentParser(description=PARSER_HELP)
    configure_parser(parser)
    args = parser.parse_args(argv)
    args.func(args)


if __name__ == "__main__":
    main(sys.argv[1:])
